# Copyright (c) 2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Parse METAR-formatted data."""
# Import the necessary libraries
from collections import namedtuple
import contextlib
from datetime import datetime, timezone
import warnings

import numpy as np
import pandas as pd

from ._metar_parser.metar_parser import parse, ParseError, TreeNode
from ._tools import open_as_needed
from .station_data import station_info
from ..package_tools import Exporter
from ..units import units

exporter = Exporter(globals())

# Configure the named tuple used for storing METAR data
Metar = namedtuple('metar', ['station_id', 'latitude', 'longitude', 'elevation', 'date_time',
                             'wind_direction', 'wind_speed', 'wind_gust', 'visibility',
                             'current_wx1', 'current_wx2', 'current_wx3', 'skyc1', 'skylev1',
                             'skyc2', 'skylev2', 'skyc3', 'skylev3', 'skyc4', 'skylev4',
                             'cloudcover', 'temperature', 'dewpoint', 'altimeter',
                             'current_wx1_symbol', 'current_wx2_symbol', 'current_wx3_symbol',
                             'remarks'])

# Create a dictionary for attaching units to the different variables
col_units = {'station_id': None,
             'latitude': 'degrees',
             'longitude': 'degrees',
             'elevation': 'meters',
             'date_time': None,
             'wind_direction': 'degrees',
             'wind_speed': 'kts',
             'wind_gust': 'kts',
             'visibility': 'meters',
             'eastward_wind': 'kts',
             'northward_wind': 'kts',
             'current_wx1': None,
             'current_wx2': None,
             'current_wx3': None,
             'low_cloud_type': None,
             'low_cloud_level': 'feet',
             'medium_cloud_type': None,
             'medium_cloud_level': 'feet',
             'high_cloud_type': None,
             'high_cloud_level': 'feet',
             'highest_cloud_type': None,
             'highest_cloud_level:': None,
             'cloud_coverage': None,
             'air_temperature': 'degC',
             'dew_point_temperature': 'degC',
             'altimeter': 'inHg',
             'air_pressure_at_sea_level': 'hPa',
             'current_wx1_symbol': None,
             'current_wx2_symbol': None,
             'current_wx3_symbol': None}


def _tree_repr_(self):
    """Produce string representation of a TreeNodex object."""
    rep = self.__class__.__name__ + '('
    args = []
    for key, value in self.__dict__.items():
        if key == 'elements':
            continue
        args.append(key + '=' + repr(value))
        dict_str = ', '.join(args) + ')'
    rep += dict_str
    return rep


# Monkey patch to improve debugging
TreeNode.__repr__ = _tree_repr_


@exporter.export
def parse_metar_to_dataframe(metar_text, *, year=None, month=None):
    """Parse a single METAR report into a Pandas DataFrame.

    Takes a METAR string in a text form, and creates a `pandas.DataFrame` including the
    essential information (not including the remarks)

    The parser follows the WMO format, allowing for missing data and assigning
    nan values where necessary. The WMO code is also provided for current weather,
    which can be utilized when plotting.

    Parameters
    ----------
    metar_text : str
        The METAR report
    year : int, optional
        Year in which observation was taken, defaults to current year. Keyword-only argument.
    month : int, optional
        Month in which observation was taken, defaults to current month. Keyword-only argument.

    Returns
    -------
    `pandas.DataFrame`

    """
    return _metars_to_dataframe([metar_text], year=year, month=month)


def parse_metar(metar_text, year, month, station_metadata=station_info):
    """Parse a METAR report in text form into a list of named tuples.

    Parameters
    ----------
    metar_text : str
        The METAR report
    station_metadata : dict
        Mapping of station identifiers to station metadata
    year : int
        Reported year of observation for constructing 'date_time'
    month : int
        Reported month of observation for constructing 'date_time'

    Returns
    -------
    metar : namedtuple
        Named tuple of parsed METAR fields

    Notes
    -----
    Returned data has named tuples with the following attributes:

    * 'station_id': Station Identifier (ex. KLOT)
    * 'latitude': Latitude of the observation, measured in degrees
    * 'longitude': Longitude of the observation, measured in degrees
    * 'elevation': Elevation of the observation above sea level, measured in meters
    * 'date_time': Date and time of the observation, datetime object
    * 'wind_direction': Direction the wind is coming from, measured in degrees
    * 'wind_speed': Wind speed, measured in knots
    * 'wind_gust': Wind gust, measured in knots
    * 'current_wx1': Current weather (1 of 3)
    * 'current_wx2': Current weather (2 of 3)
    * 'current_wx3': Current weather (3 of 3)
    * 'skyc1': Sky cover (ex. FEW)
    * 'skylev1': Height of sky cover 1, measured in feet
    * 'skyc2': Sky cover (ex. OVC)
    * 'skylev2': Height of sky cover 2, measured in feet
    * 'skyc3': Sky cover (ex. FEW)
    * 'skylev3': Height of sky cover 3, measured in feet
    * 'skyc4': Sky cover (ex. CLR)
    * 'skylev4:': Height of sky cover 4, measured in feet
    * 'cloudcover': Cloud coverage measured in oktas, taken from maximum of sky cover values
    * 'temperature': Temperature, measured in degrees Celsius
    * 'dewpoint': Dewpoint, measured in degrees Celsius
    * 'altimeter': Altimeter value, measured in inches of mercury
    * 'current_wx1_symbol': Current weather symbol (1 of 3), WMO integer code from [WMO306]_
      Attachment IV
    * 'current_wx2_symbol': Current weather symbol (2 of 3), WMO integer code from [WMO306]_
      Attachment IV
    * 'current_wx3_symbol': Current weather symbol (3 of 3), WMO integer code from [WMO306]_
      Attachment IV
    * 'visibility': Visibility distance, measured in meters
    * 'remarks': Remarks (unparsed) in the report

    """
    from ..plots.wx_symbols import wx_code_to_numeric

    # Decode the data using the parser (built using Canopy) the parser utilizes a grammar
    # file which follows the format structure dictated by the WMO Handbook, but has the
    # flexibility to decode the METAR text when there are missing or incorrectly
    # encoded values
    tree = parse(metar_text)

    # Station ID which is used to find the latitude, longitude, and elevation
    station_id = tree.siteid.text.strip()

    # Extract the latitude and longitude values from 'master' dictionary
    try:
        info = station_metadata[station_id]
        lat = info.latitude
        lon = info.longitude
        elev = info.altitude
    except KeyError:
        lat = np.nan
        lon = np.nan
        elev = np.nan

    # Set the datetime, day, and time_utc
    try:
        day_time_utc = tree.datetime.text.strip()
        day = int(day_time_utc[0:2])
        hour = int(day_time_utc[2:4])
        minute = int(day_time_utc[4:6])
        date_time = datetime(year, month, day, hour, minute)
    except ValueError:
        date_time = np.nan

    # Set the wind values
    wind_units = 'kts'
    try:
        # If there are missing wind values, set wind speed and wind direction to nan
        if ('/' in tree.wind.text) or (tree.wind.text == 'KT') or (tree.wind.text == ''):
            wind_dir = np.nan
            wind_spd = np.nan
        # If the wind direction is variable, set wind direction to nan but keep the wind speed
        else:
            wind_spd = float(tree.wind.wind_spd.text)
            if 'MPS' in tree.wind.text:
                wind_units = 'm/s'
                wind_spd = units.Quantity(wind_spd, wind_units).m_as('knots')
            if (tree.wind.wind_dir.text == 'VRB') or (tree.wind.wind_dir.text == 'VAR'):
                wind_dir = np.nan
            else:
                wind_dir = int(tree.wind.wind_dir.text)
    # If there are any errors, return nan
    except ValueError:
        wind_dir = np.nan
        wind_spd = np.nan

    # Parse out the wind gust field
    if 'G' in tree.wind.text:
        wind_gust = units.Quantity(float(tree.wind.gust.text.strip()[1:]),
                                   wind_units).m_as('knots')
    else:
        wind_gust = np.nan

    # Handle visibility
    try:
        if tree.vis.text.endswith('SM'):
            visibility = 0
            # Strip off the SM and any whitespace around the value and any leading 'M'
            vis_str = tree.vis.text[:-2].strip().lstrip('M')

            # Case of e.g. 1 1/4SM
            if ' ' in vis_str:
                whole, vis_str = vis_str.split(maxsplit=1)
                visibility += int(whole)

            # Handle fraction regardless
            if '/' in vis_str:
                num, denom = vis_str.split('/', maxsplit=1)
                if int(denom) == 0:
                    raise ValueError('Visibility denominator is 0.')
                visibility += int(num) / int(denom)
            else:  # Should be getting all cases of whole number without fraction
                visibility += int(vis_str)
            visibility = units.Quantity(visibility, 'miles').m_as('meter')
        # CAVOK means vis is "at least 10km" and no significant clouds or weather
        elif 'CAVOK' in tree.vis.text:
            visibility = 10000
        elif not tree.vis.text or tree.vis.text.strip() == '////':
            visibility = np.nan
        else:
            # Only worry about the first 4 characters (digits) and ignore possible 'NDV'
            visibility = int(tree.vis.text.strip()[:4])
    # If there are any errors, return nan
    except ValueError:
        visibility = np.nan

    # Set the weather symbols
    # If the weather symbol is missing, set values to nan
    current_wx = []
    current_wx_symbol = []
    if tree.curwx.text.strip() not in ('', '//', 'NSW'):
        current_wx = tree.curwx.text.strip().split()

        # Handle having e.g. '+' and 'TSRA' parsed into separate items
        if current_wx[0] in ('-', '+') and current_wx[1]:
            current_wx[0] += current_wx[1]
            current_wx.pop(1)

        current_wx_symbol = wx_code_to_numeric(current_wx).tolist()
    while len(current_wx) < 3:
        current_wx.append(np.nan)
    while len(current_wx_symbol) < 3:
        current_wx_symbol.append(0)

    # Set the sky conditions
    skyc = [np.nan] * 4
    skylev = [np.nan] * 4
    if tree.skyc.text[1:3] == 'VV':
        skyc[0] = 'VV'
        level = tree.skyc.text.strip()[2:5]
        skylev[0] = np.nan if not level or '/' in level else 100 * int(level)
    else:
        for ind, part in enumerate(tree.skyc.text.strip().split(maxsplit=3)):
            cover = part[:3]
            level = part[3:6]  # Strips off any ending text like in FEW017CB
            if '/' not in cover:
                skyc[ind] = cover
            if level and '/' not in level:
                with contextlib.suppress(ValueError):
                    skylev[ind] = float(level) * 100

    # Set the cloud cover variable (measured in oktas)
    if 'OVC' in tree.skyc.text or 'VV' in tree.skyc.text:
        cloudcover = 8
    elif 'BKN' in tree.skyc.text:
        cloudcover = 6
    elif 'SCT' in tree.skyc.text:
        cloudcover = 4
    elif 'FEW' in tree.skyc.text:
        cloudcover = 2
    elif ('SKC' in tree.skyc.text or 'NCD' in tree.skyc.text or 'NSC' in tree.skyc.text
          or 'CLR' in tree.skyc.text or 'CAVOK' in tree.vis.text):
        cloudcover = 0
    else:
        cloudcover = 10

    # Set the temperature and dewpoint
    temp = np.nan
    dewp = np.nan
    if tree.temp_dewp.text and tree.temp_dewp.text != ' MM/MM':
        with contextlib.suppress(ValueError):
            temp = float(tree.temp_dewp.temp.text[-2:])
            if 'M' in tree.temp_dewp.temp.text:
                temp *= -1

        with contextlib.suppress(ValueError):
            dewp = float(tree.temp_dewp.dewp.text[-2:])
            if 'M' in tree.temp_dewp.dewp.text:
                dewp *= -1

    # Set the altimeter value and sea level pressure
    if tree.altim.text:
        val = float(tree.altim.text.strip()[1:5])
        altim = val / 100 if val > 1100 else units.Quantity(val, 'hPa').m_as('inHg')
    else:
        altim = np.nan

    # Strip off extraneous stuff off the remarks section
    remarks = tree.remarks.text.lstrip().rstrip('= ')
    if remarks.startswith('RMK'):
        remarks = remarks[3:].strip()

    # Returns a named tuple with all the relevant variables
    return Metar(station_id, lat, lon, elev, date_time, wind_dir, wind_spd, wind_gust,
                 visibility, current_wx[0], current_wx[1], current_wx[2], skyc[0], skylev[0],
                 skyc[1], skylev[1], skyc[2], skylev[2], skyc[3], skylev[3], cloudcover, temp,
                 dewp, altim, current_wx_symbol[0], current_wx_symbol[1], current_wx_symbol[2],
                 remarks)


@exporter.export
def parse_metar_file(filename, *, year=None, month=None):
    """Parse a text file containing multiple METAR reports and/or text products.

    Parameters
    ----------
    filename : str or file-like object
        If str, the name of the file to be opened. If `filename` is a file-like object,
        this will be read from directly and needs to be opened in text mode (i.e. ``read()``
        needs to return a string, not bytes).
    year : int, optional
        Year in which observation was taken, defaults to current year. Keyword-only argument.
    month : int, optional
        Month in which observation was taken, defaults to current month. Keyword-only argument.

    Returns
    -------
    `pandas.DataFrame`

    """
    # Function to merge METARs
    def full_metars(x, prefix='     '):
        tmp = []
        for i in x:
            # Skip any blank lines
            if not i.strip():
                continue
            # No prefix signals a new report, so yield
            if not i.startswith(prefix) and tmp:
                yield ' '.join(tmp)
                tmp = []
            tmp.append(i.strip())

        # Handle any leftovers
        if tmp:
            yield ' '.join(tmp)

    # Open the file
    with contextlib.closing(open_as_needed(filename, 'rt')) as myfile:
        # Merge multi-line METARs into a single report--drop reports that are too short to
        # be a METAR with a robust amount of data.
        return _metars_to_dataframe(filter(lambda m: len(m) > 25, full_metars(myfile)),
                                    year=year, month=month)


def _metars_to_dataframe(metar_iter, *, year=None, month=None):
    """Turn an iterable of METAR reports into a DataFrame.

    Notes
    -----
    The output has the following columns:

    * 'station_id': Station Identifier (ex. KLOT)
    * 'latitude': Latitude of the observation, measured in degrees
    * 'longitude': Longitude of the observation, measured in degrees
    * 'elevation': Elevation of the observation above sea level, measured in meters
    * 'date_time': Date and time of the observation, datetime object
    * 'wind_direction': Direction the wind is coming from, measured in degrees
    * 'wind_speed': Wind speed, measured in knots
    * 'wind_gust': Wind gust, measured in knots
    * 'visibility': Visibility distance, measured in meters
    * 'current_wx1': Current weather (1 of 3)
    * 'current_wx2': Current weather (2 of 3)
    * 'current_wx3': Current weather (3 of 3)
    * 'low_cloud_type': Low-level sky cover (ex. FEW)
    * 'low_cloud_level': Height of low-level sky cover, measured in feet
    * 'medium_cloud_type': Medium-level sky cover (ex. OVC)
    * 'medium_cloud_level': Height of medium-level sky cover, measured in feet
    * 'high_cloud_type': High-level sky cover (ex. FEW)
    * 'high_cloud_level': Height of high-level sky cover, measured in feet
    * 'highest_cloud_type': Highest-level Sky cover (ex. CLR)
    * 'highest_cloud_level:': Height of highest-level sky cover, measured in feet
    * 'cloud_coverage': Cloud cover measured in oktas, taken from maximum of sky cover values
    * 'air_temperature': Temperature, measured in degrees Celsius
    * 'dew_point_temperature': Dew point, measured in degrees Celsius
    * 'altimeter': Altimeter value, measured in inches of mercury
    * 'remarks': Any remarks section in the report
    * 'current_wx1_symbol': Current weather symbol (1 of 3), WMO integer code from [WMO306]_
      Attachment IV
    * 'current_wx2_symbol': Current weather symbol (2 of 3), WMO integer code from [WMO306]_
      Attachment IV
    * 'current_wx3_symbol': Current weather symbol (3 of 3), WMO integer code from [WMO306]_
      Attachment IV
    * 'air_pressure_at_sea_level': Sea level pressure, derived from temperature, elevation
      and altimeter value
    * 'eastward_wind': Eastward component (u-component) of the wind vector, measured in knots
    * 'northward_wind': Northward component (v-component) of the wind vector, measured in knots

    """
    from ..calc import altimeter_to_sea_level_pressure, wind_components

    # Defaults year and/or month to present reported date if not provided
    if year is None or month is None:
        now = datetime.now(timezone.utc)
        year = now.year if year is None else year
        month = now.month if month is None else month

    # Try to parse each METAR that is given
    metars = []
    for metar in metar_iter:
        with contextlib.suppress(ParseError):
            # Parse the string of text and assign to values within the named tuple
            metars.append(parse_metar(metar, year=year, month=month))

    # Take the list of Metar objects and turn it into a DataFrame with appropriate columns
    df = pd.DataFrame(metars)
    df.set_index('station_id', inplace=True, drop=False)
    df.rename(columns={'skyc1': 'low_cloud_type', 'skylev1': 'low_cloud_level',
                       'skyc2': 'medium_cloud_type', 'skylev2': 'medium_cloud_level',
                       'skyc3': 'high_cloud_type', 'skylev3': 'high_cloud_level',
                       'skyc4': 'highest_cloud_type', 'skylev4': 'highest_cloud_level',
                       'cloudcover': 'cloud_coverage', 'temperature': 'air_temperature',
                       'dewpoint': 'dew_point_temperature'}, inplace=True)

    # Drop duplicate values
    df.drop_duplicates(subset=['date_time', 'latitude', 'longitude'], keep='last',
                       inplace=True)

    # Calculate sea-level pressure from function in metpy.calc
    df['air_pressure_at_sea_level'] = altimeter_to_sea_level_pressure(
        units.Quantity(df.altimeter.values, col_units['altimeter']),
        units.Quantity(df.elevation.values, col_units['elevation']),
        units.Quantity(df.air_temperature.values, col_units['air_temperature'])).m_as('hPa')

    # Use get wind components and assign them to eastward and northward winds
    u, v = wind_components(
        units.Quantity(df.wind_speed.values, col_units['wind_speed']),
        units.Quantity(df.wind_direction.values, col_units['wind_direction']))
    df['eastward_wind'] = u.m
    df['northward_wind'] = v.m

    # Round altimeter and sea-level pressure values
    df['altimeter'] = df.altimeter.round(2)
    df['air_pressure_at_sea_level'] = df.air_pressure_at_sea_level.round(2)

    # Set the units for the dataframe--filter out warning from Pandas
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        df.units = col_units.copy()

    return df


# Patch in the notes section into the two main parse functions
if _metars_to_dataframe.__doc__:
    # Finds the Notes snippet in the docstring
    snippet = _metars_to_dataframe.__doc__[_metars_to_dataframe.__doc__.find('Notes'):]
    parse_metar_file.__doc__ = parse_metar_file.__doc__ + snippet
    parse_metar_to_dataframe.__doc__ = parse_metar_to_dataframe.__doc__ + snippet



# Copyright (c) 2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Contains updated functions that will be modified in a future release."""

# This file left as a placeholder for function changes in the future.



# Copyright (c) 2023 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""
==================================
Using Predefined Areas with MetPy
==================================

When plotting your data on a map you want to be able to plot the data in a useful area
and with a projection that would match the areal extent well. Within MetPy we have generated
more than 400 pre-defined areas that also have associated projections, which you can use to
make quick plots over regions of interest with minimal effort on your part. While these were
intended to be used with the MetPy declarative syntax, these areas and projections are now
accessible for any use of plotting data on a map using Cartopy.
"""

from datetime import datetime, timedelta

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt

from metpy.cbook import get_test_data
from metpy.io import metar
from metpy.plots import declarative, named_areas

###########################
# Table of Predefined Areas
# -------------------------
#
# Here is a full list of all of the areas currently in MetPy with their reference name,
# descriptive name, and extent bounds.

print('area      name                bounds')
for area in named_areas:
    print(f'{named_areas[area].name:<10s}{named_areas[area].description:<20s}{named_areas[area].bounds}')

####################################################
# Example Using Bounds and Projection for an Area
# -------------------------------------------------
#
# Each area string given in the table above have a descriptive name, extent bounds, and a
# projection associated with each entry. We can pull any of this information to help us make a
# plot over the area domain.

# Select the area string
area = 'epac'

# Get the extent and project for the selected area
extent = named_areas[area].bounds
proj = named_areas[area].projection

# Plot a simple figure for the selected area
plt.figure(1, figsize=(10, 10))
ax = plt.subplot(111, projection=proj)
ax.set_extent(extent, ccrs.PlateCarree())
ax.add_feature(cfeature.STATES.with_scale('10m'), edgecolor='grey', linewidth=0.75)
ax.add_feature(cfeature.COASTLINE.with_scale('10m'), linewidth=1.1)
ax.add_feature(cfeature.BORDERS.with_scale('10m'), edgecolor='black')
ax.set_title(f'area={area}          name={named_areas[area].description}'
             f'\nproj={proj.coordinate_operation.method_name}')
plt.show()


########################################
# Examaple Using Declarative Syntax
# ---------------------------------
#
# Here is an example using a predefined area with the declarative plotting syntax for plotting
# some surface observations.

# Set the observation time
obs_time = datetime(2019, 7, 1, 12)

# Read in data
df = metar.parse_metar_file(get_test_data('metar_20190701_1200.txt', False), year=2019,
                            month=7)

# Plot desired data
obs = declarative.PlotObs()
obs.data = df
obs.time = obs_time
obs.time_window = timedelta(minutes=15)
obs.level = None
obs.fields = ['cloud_coverage', 'air_temperature', 'dew_point_temperature',
              'air_pressure_at_sea_level', 'current_wx1_symbol']
obs.plot_units = [None, 'degF', 'degF', None, None]
obs.locations = ['C', 'NW', 'SW', 'NE', 'W']
obs.formats = ['sky_cover', None, None, lambda v: format(v * 10, '.0f')[-3:],
               'current_weather']
obs.reduce_points = 0.75
obs.vector_field = ['eastward_wind', 'northward_wind']

# Panel for plot with Map features
panel = declarative.MapPanel()
panel.layout = (1, 1, 1)
panel.projection = 'area'
panel.area = 'in+'
panel.layers = ['states']
panel.title = f'Surface plot for {obs_time}'
panel.plots = [obs]

# Bringing it all together
pc = declarative.PanelContainer()
pc.size = (10, 10)
pc.panels = [panel]

pc.show()



# Copyright (c) 2017 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
r"""Tests the operation of MetPy's unit support code."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

from metpy.testing import (assert_almost_equal, assert_array_almost_equal, assert_array_equal,
                           assert_nan)
from metpy.units import (check_units, concatenate, is_quantity,
                         pandas_dataframe_to_unit_arrays, units)


def test_concatenate():
    """Test basic functionality of unit-aware concatenate."""
    result = concatenate((3 * units.meter, 400 * units.cm))
    assert_array_equal(result, np.array([3, 4]) * units.meter)
    assert not isinstance(result.m, np.ma.MaskedArray)


def test_concatenate_masked():
    """Test concatenate preserves masks."""
    d1 = units.Quantity(np.ma.array([1, 2, 3], mask=[False, True, False]), 'degC')
    result = concatenate((d1, 32 * units.degF))

    truth = np.ma.array([1, np.inf, 3, 0])
    truth[1] = np.ma.masked

    assert_array_almost_equal(result, units.Quantity(truth, 'degC'), 6)
    assert_array_equal(result.mask, np.array([False, True, False, False]))


@pytest.mark.mpl_image_compare(tolerance=0, remove_text=True)
def test_axhline():
    r"""Ensure that passing a quantity to axhline does not error."""
    fig, ax = plt.subplots()
    ax.axhline(930 * units('mbar'))
    ax.set_ylim(900, 950)
    ax.set_ylabel('')
    return fig


@pytest.mark.mpl_image_compare(tolerance=0, remove_text=True)
def test_axvline():
    r"""Ensure that passing a quantity to axvline does not error."""
    fig, ax = plt.subplots()
    ax.axvline(0 * units('degC'))
    ax.set_xlim(-1, 1)
    ax.set_xlabel('')
    return fig


#
# Tests for unit-checking decorator
#


def unit_calc(temp, press, dens, mixing, unitless_const):
    r"""Stub calculation for testing unit checking."""


test_funcs = [
    check_units('[temperature]', '[pressure]', dens='[mass]/[volume]',
                mixing='[dimensionless]')(unit_calc),
    check_units(temp='[temperature]', press='[pressure]', dens='[mass]/[volume]',
                mixing='[dimensionless]')(unit_calc),
    check_units('[temperature]', '[pressure]', '[mass]/[volume]',
                '[dimensionless]')(unit_calc)]


@pytest.mark.parametrize('func', test_funcs, ids=['some kwargs', 'all kwargs', 'all pos'])
def test_good_units(func):
    r"""Test that unit checking passes good units regardless."""
    func(30 * units.degC, 1000 * units.mbar, 1.0 * units('kg/m^3'), 1, 5.)


test_params = [((30 * units.degC, 1000 * units.mb, 1 * units('kg/m^3'), 1, 5 * units('J/kg')),
                {}, [('press', '[pressure]', 'millibarn')]),
               ((30, 1000, 1.0, 1, 5.), {}, [('press', '[pressure]', 'none'),
                                             ('temp', '[temperature]', 'none'),
                                             ('dens', '[mass]/[volume]', 'none')]),
               ((30, 1000 * units.mbar),
                {'dens': 1.0 * units('kg / m'), 'mixing': 5 * units.m, 'unitless_const': 2},
                [('temp', '[temperature]', 'none'),
                 ('dens', '[mass]/[volume]', 'kilogram / meter'),
                 ('mixing', '[dimensionless]', 'meter')])]


@pytest.mark.parametrize('func', test_funcs, ids=['some kwargs', 'all kwargs', 'all pos'])
@pytest.mark.parametrize('args,kwargs,bad_parts', test_params,
                         ids=['one bad arg', 'all args no units', 'mixed args'])
def test_bad(func, args, kwargs, bad_parts):
    r"""Test that unit checking flags appropriate arguments."""
    with pytest.raises(ValueError) as exc:
        func(*args, **kwargs)

    message = str(exc.value)
    assert func.__name__ in message
    for param in bad_parts:
        assert '`{}` requires "{}" but given "{}"'.format(*param) in message

        # Should never complain about the const argument
        assert 'unitless_const' not in message


@pytest.mark.parametrize('func', test_funcs, ids=['some kwargs', 'all kwargs', 'all pos'])
def test_bad_masked_array(func):
    """Test getting a masked array-specific message when missing units."""
    with pytest.raises(ValueError) as exc:
        func(np.ma.array([30]), 1000 * units.mbar, 1.0 * units('kg/m^3'), 1, 5.)

    message = str(exc.value)
    assert 'units.Quantity' in message


def test_pandas_units_simple():
    """Simple unit attachment to two columns."""
    df = pd.DataFrame(data=[[1, 4], [2, 5], [3, 6]], columns=['cola', 'colb'])
    df_units = {'cola': 'kilometers', 'colb': 'degC'}
    res = pandas_dataframe_to_unit_arrays(df, column_units=df_units)
    cola_truth = np.array([1, 2, 3]) * units.km
    colb_truth = np.array([4, 5, 6]) * units.degC
    assert_array_equal(res['cola'], cola_truth)
    assert_array_equal(res['colb'], colb_truth)


@pytest.mark.filterwarnings("ignore:Pandas doesn't allow columns to be created")
def test_pandas_units_on_dataframe():
    """Unit attachment based on a units attribute to a dataframe."""
    df = pd.DataFrame(data=[[1, 4], [2, 5], [3, 6]], columns=['cola', 'colb'])
    df.units = {'cola': 'kilometers', 'colb': 'degC'}
    res = pandas_dataframe_to_unit_arrays(df)
    cola_truth = np.array([1, 2, 3]) * units.km
    colb_truth = np.array([4, 5, 6]) * units.degC
    assert_array_equal(res['cola'], cola_truth)
    assert_array_equal(res['colb'], colb_truth)


@pytest.mark.filterwarnings("ignore:Pandas doesn't allow columns to be created")
def test_pandas_units_on_dataframe_not_all_with_units():
    """Unit attachment with units attribute with a column with no units."""
    df = pd.DataFrame(data=[[1, 4], [2, 5], [3, 6]], columns=['cola', 'colb'])
    df.units = {'cola': 'kilometers'}
    res = pandas_dataframe_to_unit_arrays(df)
    cola_truth = np.array([1, 2, 3]) * units.km
    colb_truth = np.array([4, 5, 6])
    assert_array_equal(res['cola'], cola_truth)
    assert_array_equal(res['colb'], colb_truth)


def test_pandas_units_no_units_given():
    """Ensure unit attachment fails if no unit information is given."""
    df = pd.DataFrame(data=[[1, 4], [2, 5], [3, 6]], columns=['cola', 'colb'])
    with pytest.raises(ValueError):
        pandas_dataframe_to_unit_arrays(df)


def test_added_degrees_units():
    """Test that our added degrees units are present in the registry."""
    # Test equivalence of abbreviations/aliases to our defined names
    assert str(units('degrees_N').units) == 'degrees_north'
    assert str(units('degreesN').units) == 'degrees_north'
    assert str(units('degree_north').units) == 'degrees_north'
    assert str(units('degree_N').units) == 'degrees_north'
    assert str(units('degreeN').units) == 'degrees_north'
    assert str(units('degrees_E').units) == 'degrees_east'
    assert str(units('degreesE').units) == 'degrees_east'
    assert str(units('degree_east').units) == 'degrees_east'
    assert str(units('degree_E').units) == 'degrees_east'
    assert str(units('degreeE').units) == 'degrees_east'

    # Test equivalence of our defined units to base units
    assert units('degrees_north') == units('degrees')
    assert units('degrees_north').to_base_units().units == units.radian
    assert units('degrees_east') == units('degrees')
    assert units('degrees_east').to_base_units().units == units.radian
    assert_almost_equal(0 * units.dBz, 1 * units('mm^6/m^3'))


def test_is_quantity():
    """Test is_quantity properly works."""
    assert is_quantity(1 * units.m)
    assert not is_quantity(np.array([1]))


def test_is_quantity_multiple():
    """Test is_quantity with multiple inputs."""
    assert is_quantity(1 * units.m, np.array([4.]) * units.degree)
    assert not is_quantity(1 * units.second, np.array([5., 2.]))


def test_gpm_unit():
    """Test that the gpm unit does alias to meters."""
    x = 1 * units('gpm')
    assert x.units == units('meter')


def test_assert_nan():
    """Test that assert_nan actually fails when not given a NaN."""
    with pytest.raises(AssertionError):
        assert_nan(1.0 * units.m, units.inches)


def test_assert_nan_checks_units():
    """Test that assert_nan properly checks units."""
    with pytest.raises(AssertionError):
        assert_nan(np.nan * units.m, units.second)


def test_percent_units():
    """Test that percent sign units are properly parsed and interpreted."""
    assert units('%').units == units('percent')


@pytest.mark.parametrize(
    'unit_str,pint_unit',
    (
        # Validated against cf-units (UDUNITS-2 wrapper)
        ('m s-1', units.m / units.s),
        ('m2 s-2', units.m ** 2 / units.s ** 2),
        ('kg(-1)', units.kg),
        ('kg-1', units.kg ** -1),
        ('W m(-2)', units.m ** 3 * units.kg * units.s ** -3),
        ('W m-2', units.kg * units.s ** -3),
        ('(W m-2 um-1)-1', units.m * units.kg ** -1 * units.s ** 3),
        pytest.param(
            '(J kg-1)(m s-1)-1', units.m / units.s,
            marks=pytest.mark.xfail(reason='hgrecco/pint#1485')
        ),
        ('(J kg-1)(m s-1)(-1)', units.m ** 3 / units.s ** 3),
        ('/s', units.s ** -1)
    )
)
def test_udunits_power_syntax(unit_str, pint_unit):
    """Test that UDUNITS style powers are properly parsed and interpreted."""
    assert units(unit_str).to_base_units().units == pint_unit



# Copyright (c) 2018,2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tools and calculations for interpolating specifically to a grid."""

import numpy as np

from .points import (interpolate_to_points, inverse_distance_to_points,
                     natural_neighbor_to_points)
from ..package_tools import Exporter
from ..pandas import preprocess_pandas

exporter = Exporter(globals())


def generate_grid(horiz_dim, bbox):
    r"""Generate a meshgrid based on bounding box and x & y resolution.

    Parameters
    ----------
    horiz_dim: int or float
        Horizontal resolution
    bbox: dict
        Dictionary with keys 'east', 'west', 'north', 'south' with the box extents
        in those directions.

    Returns
    -------
    grid_x: (X, Y) numpy.ndarray
        X dimension meshgrid defined by given bounding box
    grid_y: (X, Y) numpy.ndarray
        Y dimension meshgrid defined by given bounding box

    """
    x_steps, y_steps = get_xy_steps(bbox, horiz_dim)

    grid_x = np.linspace(bbox['west'], bbox['east'], x_steps)
    grid_y = np.linspace(bbox['south'], bbox['north'], y_steps)

    return np.meshgrid(grid_x, grid_y)


def generate_grid_coords(gx, gy):
    r"""Calculate x,y coordinates of each grid cell.

    Parameters
    ----------
    gx: numeric
        x coordinates in meshgrid
    gy: numeric
        y coordinates in meshgrid

    Returns
    -------
    (X, Y) numpy.ndarray
        List of coordinates in meshgrid

    """
    return np.stack([gx.ravel(), gy.ravel()], axis=1)


def get_xy_range(bbox):
    r"""Return x and y ranges in meters based on bounding box.

    bbox: dict
        Dictionary with keys 'east', 'west', 'north', 'south' with the box extents
        in those directions.

    Returns
    -------
    x_range: float
        Range in meters in x dimension.
    y_range: float
        Range in meters in y dimension.

    """
    x_range = bbox['east'] - bbox['west']
    y_range = bbox['north'] - bbox['south']

    return x_range, y_range


def get_xy_steps(bbox, h_dim):
    r"""Return meshgrid spacing based on bounding box.

    bbox: dict
        Dictionary with keys 'east', 'west', 'north', 'south' with the box extents
        in those directions.
    h_dim: int or float
        Horizontal resolution in meters.

    Returns
    -------
    x_steps, (X, ) numpy.ndarray
        Number of grids in x dimension.
    y_steps: (Y, ) numpy.ndarray
        Number of grids in y dimension.

    """
    x_range, y_range = get_xy_range(bbox)

    x_steps = np.ceil(x_range / h_dim) + 1
    y_steps = np.ceil(y_range / h_dim) + 1

    return int(x_steps), int(y_steps)


def get_boundary_coords(x, y, spatial_pad=0):
    r"""Return bounding box based on given x and y coordinates assuming northern hemisphere.

    x: numeric
        x coordinates.
    y: numeric
        y coordinates.
    spatial_pad: int or float
        Number of meters to add to the x and y dimensions to reduce edge effects.

    Returns
    -------
    bbox: dict
        Dictionary with keys 'east', 'west', 'north', 'south' with the box extents
        in those directions.

    """
    west = np.min(x) - spatial_pad
    east = np.max(x) + spatial_pad
    north = np.max(y) + spatial_pad
    south = np.min(y) - spatial_pad

    return {'west': west, 'south': south, 'east': east, 'north': north}


@exporter.export
def natural_neighbor_to_grid(xp, yp, variable, grid_x, grid_y):
    r"""Generate a natural neighbor interpolation of the given points to a regular grid.

    This assigns values to the given grid using the Liang and Hale [Liang2010]_.
    approach.

    Parameters
    ----------
    xp: (P, ) numpy.ndarray
        x-coordinates of observations
    yp: (P, ) numpy.ndarray
        y-coordinates of observations
    variable: (P, ) numpy.ndarray
        observation values associated with (xp, yp) pairs.
        IE, variable[i] is a unique observation at (xp[i], yp[i])
    grid_x: (M, N) numpy.ndarray
        Meshgrid associated with x dimension
    grid_y: (M, N) numpy.ndarray
        Meshgrid associated with y dimension

    Returns
    -------
    img: (M, N) numpy.ndarray
        Interpolated values on a 2-dimensional grid

    See Also
    --------
    natural_neighbor_to_points

    """
    # Handle grid-to-points conversion, and use function from `interpolation`
    points_obs = list(zip(xp, yp))
    points_grid = generate_grid_coords(grid_x, grid_y)
    img = natural_neighbor_to_points(points_obs, variable, points_grid)
    return img.reshape(grid_x.shape)


@exporter.export
def inverse_distance_to_grid(xp, yp, variable, grid_x, grid_y, r, gamma=None, kappa=None,
                             min_neighbors=3, kind='cressman'):
    r"""Generate an inverse distance interpolation of the given points to a regular grid.

    Values are assigned to the given grid using inverse distance weighting based on either
    [Cressman1959]_ or [Barnes1964]_. The Barnes implementation used here based on [Koch1983]_.

    Parameters
    ----------
    xp: (N, ) numpy.ndarray
        x-coordinates of observations.
    yp: (N, ) numpy.ndarray
        y-coordinates of observations.
    variable: (N, ) numpy.ndarray
        observation values associated with (xp, yp) pairs.
        IE, variable[i] is a unique observation at (xp[i], yp[i]).
    grid_x: (M, 2) numpy.ndarray
        Meshgrid associated with x dimension.
    grid_y: (M, 2) numpy.ndarray
        Meshgrid associated with y dimension.
    r: float
        Radius from grid center, within which observations
        are considered and weighted.
    gamma: float
        Adjustable smoothing parameter for the barnes interpolation. Default None.
    kappa: float
        Response parameter for barnes interpolation. Default None.
    min_neighbors: int
        Minimum number of neighbors needed to perform barnes or cressman interpolation
        for a point. Default is 3.
    kind: str
        Specify what inverse distance weighting interpolation to use.
        Options: 'cressman' or 'barnes'. Default 'cressman'

    Returns
    -------
    img: (M, N) numpy.ndarray
        Interpolated values on a 2-dimensional grid

    See Also
    --------
    inverse_distance_to_points

    """
    # Handle grid-to-points conversion, and use function from `interpolation`
    points_obs = list(zip(xp, yp))
    points_grid = generate_grid_coords(grid_x, grid_y)
    img = inverse_distance_to_points(points_obs, variable, points_grid, r, gamma=gamma,
                                     kappa=kappa, min_neighbors=min_neighbors, kind=kind)
    return img.reshape(grid_x.shape)


@exporter.export
@preprocess_pandas
def interpolate_to_grid(x, y, z, interp_type='linear', hres=50000,
                        minimum_neighbors=3, gamma=0.25, kappa_star=5.052,
                        search_radius=None, rbf_func='linear', rbf_smooth=0,
                        boundary_coords=None):
    r"""Interpolate given (x,y), observation (z) pairs to a grid based on given parameters.

    Parameters
    ----------
    x: array-like
        x coordinate, can have units of linear distance or degrees
    y: array-like
        y coordinate, can have units of linear distance or degrees
    z: array-like
        observation value
    interp_type: str
        What type of interpolation to use. Available options include:
        1) "linear", "nearest", "cubic", or "rbf" from `scipy.interpolate`.
        2) "natural_neighbor", "barnes", or "cressman" from `metpy.interpolate`.
        Default "linear".
    hres: float
        The horizontal resolution of the generated grid, given in the same units as the
        x and y parameters. Default 50000.
    minimum_neighbors: int
        Minimum number of neighbors needed to perform Barnes or Cressman interpolation for a
        point. Default is 3.
    gamma: float
        Adjustable smoothing parameter for the barnes interpolation. Default 0.25.
    kappa_star: float
        Response parameter for barnes interpolation, specified nondimensionally
        in terms of the Nyquist. Default 5.052
    search_radius: float
        A search radius to use for the Barnes and Cressman interpolation schemes.
        If search_radius is not specified, it will default to 5 times the average spacing of
        observations.
    rbf_func: str
        Specifies which function to use for Rbf interpolation.
        Options include: 'multiquadric', 'inverse', 'gaussian', 'linear', 'cubic',
        'quintic', and 'thin_plate'. Default 'linear'. See `scipy.interpolate.Rbf` for more
        information.
    rbf_smooth: float
        Smoothing value applied to rbf interpolation.  Higher values result in more smoothing.
    boundary_coords: dict
        Optional dictionary containing coordinates of the study area boundary. Dictionary
        should be in format: {'west': west, 'south': south, 'east': east, 'north': north}

    Returns
    -------
    grid_x: (N, 2) numpy.ndarray
        Meshgrid for the resulting interpolation in the x dimension
    grid_y: (N, 2) numpy.ndarray
        Meshgrid for the resulting interpolation in the y dimension numpy.ndarray
    img: (M, N) numpy.ndarray
        2-dimensional array representing the interpolated values for each grid.

    See Also
    --------
    interpolate_to_points

    Notes
    -----
    This function acts as a wrapper for `interpolate_points` to allow it to generate a regular
    grid.

    This function interpolates points to a Cartesian plane, even if lat/lon coordinates
    are provided.

    """
    # Generate the grid
    if boundary_coords is None:
        boundary_coords = get_boundary_coords(x, y)
    grid_x, grid_y = generate_grid(hres, boundary_coords)

    # Handle grid-to-points conversion, and use function from `interpolation`
    points_obs = np.array(list(zip(x, y)))
    points_grid = generate_grid_coords(grid_x, grid_y)
    img = interpolate_to_points(points_obs, z, points_grid, interp_type=interp_type,
                                minimum_neighbors=minimum_neighbors, gamma=gamma,
                                kappa_star=kappa_star, search_radius=search_radius,
                                rbf_func=rbf_func, rbf_smooth=rbf_smooth)

    return grid_x, grid_y, img.reshape(grid_x.shape)


@exporter.export
@preprocess_pandas
def interpolate_to_isosurface(level_var, interp_var, level, bottom_up_search=True):
    r"""Linear interpolation of a variable to a given vertical level from given values.

    This function assumes that highest vertical level (lowest pressure) is zeroth index.
    A classic use of this function would be to compute the potential temperature on the
    dynamic tropopause (2 PVU surface).

    Parameters
    ----------
    level_var: array-like (P, M, N)
        Level values in 3D grid on common vertical coordinate (e.g., PV values on
        isobaric levels). Assumes height dimension is highest to lowest in atmosphere.
    interp_var: array-like (P, M, N)
        Variable on 3D grid with same vertical coordinate as level_var to interpolate to
        given level (e.g., potential temperature on isobaric levels)
    level: int or float
        Desired interpolated level (e.g., 2 PVU surface)
    bottom_up_search : bool, optional
        Controls whether to search for levels bottom-up (starting at lower indices),
        or top-down (starting at higher indices). Defaults to True, which is bottom-up search.

    Returns
    -------
    interp_level: (M, N) numpy.ndarray
        The interpolated variable (e.g., potential temperature) on the desired level (e.g.,
        2 PVU surface)

    Notes
    -----
    This function implements a linear interpolation to estimate values on a given surface.
    The prototypical example is interpolation of potential temperature to the dynamic
    tropopause (e.g., 2 PVU surface)

    """
    from ..calc import find_bounding_indices

    # Find index values above and below desired interpolated surface value
    above, below, good = find_bounding_indices(level_var, [level], axis=0,
                                               from_below=bottom_up_search)

    # Linear interpolation of variable to interpolated surface value
    interp_level = (((level - level_var[above]) / (level_var[below] - level_var[above]))
                    * (interp_var[below] - interp_var[above])) + interp_var[above]

    # Handle missing values and instances where no values for surface exist above and below
    interp_level[~good] = np.nan
    minvar = (np.min(level_var, axis=0) >= level)
    maxvar = (np.max(level_var, axis=0) <= level)
    interp_level[0][minvar] = interp_var[-1][minvar]
    interp_level[0][maxvar] = interp_var[0][maxvar]
    return interp_level.squeeze()



# Copyright (c) 2022 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Collection of pre-defined areas for plotting with associated projections."""

from collections import namedtuple

Area = namedtuple('Area', ['name', 'description', 'bounds', 'projection'])

try:
    import cartopy.crs as ccrs

    area_data = [
        Area('us', 'UNITED STATES', (-119.0, -56.0, 19.0, 47.0),
             ccrs.NorthPolarStereo(central_longitude=-100)),
        Area('awips', 'AWIPS US', (-127.0, -59.0, 20.0, 50.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('uslcc', 'UNITED STATES LCC', (-124.0, -62.0, 20.0, 51.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('whlf', 'WESTERN HALF', (-123.8, -85.9, 22.9, 50.2),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-105)),
        Area('chlf', 'CENTRAL HALF', (-111.0, -79.0, 27.5, 50.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('centus', 'CENTRAL US', (-105.4, -77.0, 24.7, 47.6),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('ehlf', 'EASTERN HALF', (-96.2, -62.7, 22.0, 49.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-90)),
        Area('mehlf', 'EASTERN HALF', (-89.9, -66.6, 23.8, 49.1),
             ccrs.PlateCarree()),
        Area('west', 'WESTERN US', (-125.0, -90.0, 25.0, 55.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-105)),
        Area('cent', 'CENTRAL US', (-107.4, -75.3, 24.3, 49.7),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-96)),
        Area('east', 'EASTERN US', (-100.55, -65.42, 24.57, 47.2),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-90)),
        Area('nwus', 'NW SECTOR', (-126.0, -102.0, 38.25, 50.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-114)),
        Area('swus', 'SW SECTOR', (-126.0, -100.0, 28.25, 40.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-110)),
        Area('ncus', 'NC SECTOR', (-108.0, -84.0, 38.25, 50.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-98)),
        Area('scus', 'SC SECTOR', (-108.9, -84.0, 24.0, 40.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('neus', 'NE SECTOR', (-89.0, -64.0, 37.25, 47.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-78)),
        Area('seus', 'SE SECTOR', (-90.0, -66.0, 28.25, 40.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-87)),
        Area('bosfa', 'BOSTON FA AREA', (-87.5, -63.5, 34.5, 50.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-75)),
        Area('miafa', 'MIAMI FA AREA', (-88.0, -72.0, 23.0, 39.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-80)),
        Area('chifa', 'CHICAGO FA AREA', (-108.0, -75.0, 34.0, 50.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-93)),
        Area('dfwfa', 'DALLAS FA AREA', (-106.5, -80.5, 22.0, 40.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-99)),
        Area('slcfa', 'SALT LAKE FA AREA', (-126.0, -98.0, 29.5, 50.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-109)),
        Area('sfofa', 'SAN FRAN FA AREA', (-129.0, -111.0, 30.0, 50.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-118)),
        Area('ak', 'ALASKA HPC QPF', (-180.0, -106.0, 42.0, 73.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-143)),
        Area('al', 'ALABAMA', (-95.0, -79.0, 27.0, 38.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-87)),
        Area('ar', 'ARKANSAS', (-100.75, -84.75, 29.5, 40.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-93)),
        Area('ca', 'CALIFORNIA', (-127.75, -111.75, 31.5, 42.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-120)),
        Area('co', 'COLORADO', (-114.0, -98.0, 33.5, 44.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-106)),
        Area('ct', 'CONNECTICUT', (-81.25, -65.25, 36.0, 47.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-73)),
        Area('dc', 'WASHINGTON DC', (-85.0, -69.0, 33.35, 44.35),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-77)),
        Area('de', 'DELAWARE', (-83.75, -67.75, 33.25, 44.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-76)),
        Area('fl', 'FLORIDA', (-90.0, -74.0, 23.0, 34.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-82)),
        Area('ga', 'GEORGIA', (-92.0, -76.0, 27.5, 38.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-84)),
        Area('hi', 'HAWAII', (-161.5, -152.5, 17.0, 23.0),
             ccrs.PlateCarree()),
        Area('ia', 'IOWA', (-102.0, -86.0, 36.5, 47.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-94)),
        Area('id', 'IDAHO', (-123.0, -107.0, 39.25, 50.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-115)),
        Area('il', 'ILLINOIS', (-97.75, -81.75, 34.5, 45.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-90)),
        Area('in', 'INDIANA', (-94.5, -78.5, 34.5, 45.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-87)),
        Area('ks', 'KANSAS', (-106.5, -90.5, 33.25, 44.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-99)),
        Area('ky', 'KENTUCKY', (-93.0, -77.0, 31.75, 42.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-85)),
        Area('la', 'LOUISIANA', (-100.75, -84.75, 25.75, 36.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-93)),
        Area('ma', 'MASSACHUSETTS', (-80.25, -64.25, 36.75, 47.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-72)),
        Area('md', 'MARYLAND', (-85.25, -69.25, 33.75, 44.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-77)),
        Area('me', 'MAINE', (-77.75, -61.75, 39.5, 50.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-70)),
        Area('mi', 'MICHIGAN', (-93.0, -77.0, 37.75, 48.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-85)),
        Area('mn', 'MINNESOTA', (-102.0, -86.0, 40.5, 51.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-94)),
        Area('mo', 'MISSOURI', (-101.0, -85.0, 33.0, 44.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-93)),
        Area('ms', 'MISSISSIPPI', (-98.0, -82.0, 27.0, 38.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-90)),
        Area('mt', 'MONTANA', (-117.0, -101.0, 41.5, 52.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-109)),
        Area('nc', 'NORTH CAROLINA', (-87.25, -71.25, 30.0, 41.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-79)),
        Area('nd', 'NORTH DAKOTA', (-107.5, -91.5, 42.25, 53.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('ne', 'NEBRASKA', (-107.5, -91.5, 36.25, 47.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('nh', 'NEW HAMPSHIRE', (-79.5, -63.5, 38.25, 49.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-72)),
        Area('nj', 'NEW JERSEY', (-82.5, -66.5, 34.75, 45.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-75)),
        Area('nm', 'NEW MEXICO', (-114.25, -98.25, 29.0, 40.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-106)),
        Area('nv', 'NEVADA', (-125.0, -109.0, 34.0, 45.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-117)),
        Area('ny', 'NEW YORK', (-84.0, -68.0, 37.25, 48.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-76)),
        Area('oh', 'OHIO', (-91.0, -75.0, 34.5, 45.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-83)),
        Area('ok', 'OKLAHOMA', (-105.25, -89.25, 30.25, 41.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-97)),
        Area('or', 'OREGON', (-128.0, -112.0, 38.75, 49.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-120)),
        Area('pa', 'PENNSYLVANIA', (-86.0, -70.0, 35.5, 46.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-78)),
        Area('ri', 'RHODE ISLAND', (-79.75, -63.75, 36.0, 47.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-72)),
        Area('sc', 'SOUTH CAROLINA', (-89.0, -73.0, 28.5, 39.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-81)),
        Area('sd', 'SOUTH DAKOTA', (-107.5, -91.5, 39.0, 50.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('tn', 'TENNESSEE', (-95.0, -79.0, 30.0, 41.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-87)),
        Area('tx', 'TEXAS', (-107.0, -91.0, 25.4, 36.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-99)),
        Area('ut', 'UTAH', (-119.0, -103.0, 34.0, 45.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-111)),
        Area('va', 'VIRGINIA', (-86.5, -70.5, 32.25, 43.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-79)),
        Area('vt', 'VERMONT', (-80.75, -64.75, 38.25, 49.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-73)),
        Area('wi', 'WISCONSIN', (-98.0, -82.0, 38.5, 49.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-90)),
        Area('wv', 'WEST VIRGINIA', (-89.0, -73.0, 33.0, 44.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-81)),
        Area('wy', 'WYOMING', (-116.0, -100.0, 37.75, 48.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-108)),
        Area('az', 'ARIZONA', (-119.0, -103.0, 29.0, 40.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-111)),
        Area('wa', 'WASHINGTON', (-128.0, -112.0, 41.75, 52.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-120)),
        Area('cn', 'CANADA', (-120.4, -14.0, 37.9, 58.6),
             ccrs.NorthPolarStereo(central_longitude=-90)),
        Area('cnlcc', 'CANADA LCC', (-127, -54.0, 40, 70),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-100)),
        Area('ab', 'ALBERTA', (-119.6, -108.2, 48.6, 60.4),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-114)),
        Area('bc', 'BRITISH COLUMBIA', (-134.5, -109.0, 47.2, 60.7),
             ccrs.NorthPolarStereo(central_longitude=-122)),
        Area('mb', 'MANITOBA', (-102.4, -86.1, 48.3, 60.2),
             ccrs.NorthPolarStereo(central_longitude=-95)),
        Area('nb', 'NEW BRUNSWICK', (-75.7, -57.6, 42.7, 49.6),
             ccrs.NorthPolarStereo(central_longitude=-67)),
        Area('nf', 'NEWFOUNDLAND', (-68.0, -47.0, 45.0, 62.0),
             ccrs.NorthPolarStereo(central_longitude=-58)),
        Area('ns', 'NOVA SCOTIA', (-67.0, -59.0, 43.0, 47.5),
             ccrs.NorthPolarStereo(central_longitude=-63)),
        Area('nt', 'NW TERRITORIES', (-132.0, -100.0, 59.0, 70.5),
             ccrs.NorthPolarStereo(central_longitude=-112)),
        Area('nu', 'NUNAVUT', (-115.0, -65.0, 59.0, 80),
             ccrs.NorthPolarStereo(central_longitude=-95)),
        Area('on', 'ONTARIO', (-94.5, -68.2, 41.9, 55.0),
             ccrs.NorthPolarStereo(central_longitude=-90)),
        Area('pe', 'PRINCE EDWARD IS', (-64.6, -61.7, 45.8, 47.1),
             ccrs.NorthPolarStereo(central_longitude=-63)),
        Area('qb', 'QUEBEC', (-80.0, -49.2, 44.1, 60.9),
             ccrs.NorthPolarStereo(central_longitude=-74)),
        Area('sa', 'SASKATCHEWAN', (-111.2, -97.8, 48.5, 60.3),
             ccrs.NorthPolarStereo(central_longitude=-105)),
        Area('yt', 'YUKON TERRITORY', (-142.0, -117.0, 59.0, 70.5),
             ccrs.NorthPolarStereo(central_longitude=-130)),
        Area('tropatl', 'TROPICAL ATLANTIC', (-115.0, 10.0, 0.0, 40.0),
             ccrs.PlateCarree()),
        Area('subtrop', 'SUBTROPICAL ATL', (-90.0, -20.0, 20.0, 60.0),
             ccrs.PlateCarree()),
        Area('troppac', 'TROPICAL PACIFIC', (-165.0, -80.0, -25.0, 45.0),
             ccrs.PlateCarree()),
        Area('gulf', 'GULF OF MEXICO', (-105.0, -70.0, 10.0, 40.0),
             ccrs.PlateCarree()),
        Area('carib', 'CARIBBEAN SEA', (-100.0, -50.0, 0.0, 40.0),
             ccrs.PlateCarree()),
        Area('sthepac', 'SOUTH E PACIFIC', (-170.0, -70.0, -60.0, 0.0),
             ccrs.PlateCarree()),
        Area('natlmed', 'N ATL/MEDITER', (-30.0, 70.0, 0.0, 65.0),
             ccrs.PlateCarree()),
        Area('atlhur', 'ATLANTIC HURR', (-96.0, -6.0, 4.0, 3.0),
             ccrs.NorthPolarStereo(central_longitude=-90)),
        Area('nam', 'N AMERICA', (-144, -50, 12, 75),
             ccrs.NorthPolarStereo(central_longitude=-105)),
        Area('sam', 'S AMERICA', (-120.0, -20.0, -60.0, 20.0),
             ccrs.PlateCarree()),
        Area('samps', 'S AMERICA (PS)', (-100.0, -34.0, -58.0, 15.0),
             ccrs.RotatedPole(pole_latitude=90, pole_longitude=0,
                              central_rotated_longitude=-70)),
        Area('eur', 'EUROPE', (-16.0, 80.0, 24.0, 52.0),
             ccrs.NorthPolarStereo(central_longitude=15)),
        Area('natl', 'N ATLANTIC', (-110.0, 20.1, 15.0, 70.0),
             ccrs.PlateCarree()),
        Area('watl', 'W ATLANTIC', (-84.0, -38.0, 25.0, 46.0),
             ccrs.NorthPolarStereo(central_longitude=-67)),
        Area('tatl', 'TROPICAL ATLANTIC', (-90.0, -15.0, -10.0, 35.0),
             ccrs.PlateCarree()),
        Area('npac', 'N PACIFIC', (102.0, 250.0, -12.0, 60.0),
             ccrs.PlateCarree(central_longitude=-180)),
        Area('spac', 'S PACIFIC', (102.0, 290.0, -60.0, 20.0),
             ccrs.PlateCarree(central_longitude=-180)),
        Area('tpac', 'TROPICAL PACIFIC', (-165.0, -75.0, -10.0, 40.0),
             ccrs.PlateCarree()),
        Area('epac', 'E PACIFIC', (-150.0, -98.0, 12.0, 60.0),
             ccrs.RotatedPole(pole_latitude=90, pole_longitude=0,
                              central_rotated_longitude=-130)),
        Area('wpac', 'W PACIFIC', (130.0, 240.0, 0.0, 63.0),
             ccrs.PlateCarree(central_longitude=-180)),
        Area('mpac', 'MARINE PACIFIC', (128.0, 252.0, 15.0, 71.95),
             ccrs.PlateCarree(central_longitude=-180)),
        Area('americas', 'ICAO MERC AREA A', (-137.4, -12.6, -54.0, 67.0),
             ccrs.PlateCarree()),
        Area('atlantic', 'ICAO MERC AREA B1', (-125.0, 40.0, -45.5, 62.7),
             ccrs.PlateCarree()),
        Area('africa', 'ICAO MERC AREA C', (-35.0, 70.0, -45.0, 75.0),
             ccrs.PlateCarree()),
        Area('indocean', 'ICAO MERC AREA D', (-15.0, 132.0, -27.0, 63.0),
             ccrs.PlateCarree()),
        Area('oceania', 'ICAO MERC AREA E', (25.0, 180.0, -54.0, 40.0),
             ccrs.PlateCarree()),
        Area('trop_pac', 'ICAO MERC AREA F', (100.0, 250, -52.7, 50.0),
             ccrs.PlateCarree()),
        Area('asia_ps', 'ICAO P.S. AREA G', (34.8, 157.2, -0.8, 13.7),
             ccrs.NorthPolarStereo(central_longitude=65)),
        Area('na_ps', 'ICAO P.S. AREA H', (-79.1, 56.7, 1.6, 25.2),
             ccrs.NorthPolarStereo(central_longitude=-45)),
        Area('npac_ps', 'ICAO P.S. AREA I', (166.24, -60.62, -6.74, 33.32),
             ccrs.NorthPolarStereo(central_longitude=-155)),
        Area('antarctica', 'ICAO P.S. AREA J', (106.8, -101.1, -27.6, 0.8),
             ccrs.SouthPolarStereo(central_longitude=-142)),
        Area('europe', 'ICAO P.S. AREA EU', (-21.6, 68.4, 21.4, 58.7),
             ccrs.Stereographic(central_latitude=90, central_longitude=11)),
        Area('middle_east', 'ICAO MERC AREA ME', (17.0, 70.0, 10.0, 44.0),
             ccrs.PlateCarree()),
        Area('india', 'ICAO MERC AREA AS', (53.0, 108.0, 0.0, 36.0),
             ccrs.PlateCarree()),
        Area('ag', 'ARGENTINA', (-80.0, -53.0, -56.0, -20.0),
             ccrs.SouthPolarStereo(central_longitude=-67)),
        Area('ah', 'AFGHANISTAN', (60.0, 77.0, 27.0, 40.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=69)),
        Area('afrca', 'AFRICA', (-25.0, 59.4, -36.0, 41.0),
             ccrs.PlateCarree()),
        Area('ai', 'ASCENSION ISLAND', (-14.3, -14.1, -8.0, -7.8),
             ccrs.PlateCarree()),
        Area('alba', 'ALBANIA', (18.0, 23.0, 39.0, 43.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=21)),
        Area('alge', 'ALGERIA', (-9.0, 12.0, 15.0, 38.0),
             ccrs.PlateCarree()),
        Area('an', 'ANGOLA', (10.0, 25.0, -20.0, -5.0),
             ccrs.PlateCarree()),
        Area('antl', 'LESSER ANTILLES', (-70.0, -58.0, 11.0, 19.0),
             ccrs.PlateCarree()),
        Area('antg', 'GREATER ANTILLES', (-86.0, -65.0, 17.0, 25.0),
             ccrs.PlateCarree()),
        Area('atg', 'ANTIGUA & BARBUDA', (-62.0, -61.6, 16.9, 17.75),
             ccrs.PlateCarree()),
        Area('au', 'AUSTRALIA', (110.0, 160.0, -40.0, -6.5),
             ccrs.SouthPolarStereo(central_longitude=140)),
        Area('azor', 'AZORES', (-27.6, -23.0, 36.0, 41.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-25)),
        Area('ba', 'BAHAMAS', (-80.5, -72.5, 22.5, 28.5),
             ccrs.PlateCarree()),
        Area('be', 'BERMUDA', (-64.9, -64.5, 32.2, 32.6),
             ccrs.PlateCarree()),
        Area('bel', 'BELGIUM', (2.5, 6.5, 49.4, 51.6),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=5)),
        Area('bf', 'BRUNEI', (113.0, 116.0, 4.0, 5.5),
             ccrs.PlateCarree()),
        Area('bfa', 'BURKINA FASO', (-6.0, 3.0, 9.0, 15.1),
             ccrs.PlateCarree()),
        Area('bh', 'BELIZE', (-89.3, -88.1, 15.7, 18.5),
             ccrs.PlateCarree()),
        Area('bi', 'BURUNDI', (29.0, 30.9, -4.6, -2.2),
             ccrs.PlateCarree()),
        Area('bj', 'BENIN', (0.0, 5.0, 6.0, 12.6),
             ccrs.PlateCarree()),
        Area('bn', 'BAHRAIN', (50.0, 51.0, 25.5, 27.1),
             ccrs.PlateCarree()),
        Area('bo', 'BOLIVIA', (-72.0, -50.0, -24.0, -8.0),
             ccrs.PlateCarree()),
        Area('bots', 'BOTSWANA', (19.0, 29.6, -27.0, -17.0),
             ccrs.PlateCarree()),
        Area('br', 'BARBADOS', (-62.5, -56.5, 12.45, 13.85),
             ccrs.PlateCarree()),
        Area('bt', 'BRITISH INDIAN OC', (71.25, 72.6, -7.5, -5.0),
             ccrs.PlateCarree()),
        Area('bu', 'BULGARIA', (22.0, 30.0, 40.0, 45.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=26)),
        Area('bv', 'BOUVET ISLANDS', (3.0, 4.0, -55.0, -54.0),
             ccrs.SouthPolarStereo(central_longitude=4)),
        Area('bw', 'BANGLADESH', (87.0, 93.0, 20.8, 27.0),
             ccrs.PlateCarree()),
        Area('by', 'BYELORUSSIA', (19.0, 33.0, 51.0, 60.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=26)),
        Area('bz', 'BRAZIL', (-75.0, -30.0, -35.0, 5.0),
             ccrs.PlateCarree()),
        Area('cais', 'CANTON ISLAND', (-172.0, -171.0, -3.0, -2.0),
             ccrs.PlateCarree()),
        Area('nwcar', 'NWX CARIBBEAN', (-120.0, -50.0, -15.0, 35.0),
             ccrs.PlateCarree()),
        Area('cari', 'CARIBBEAN', (-103.0, -53.0, 3.0, 36.0),
             ccrs.PlateCarree()),
        Area('cb', 'CHAD', (13.0, 25.0, 7.0, 24.0),
             ccrs.PlateCarree()),
        Area('ce', 'CENTRAL AFRICA', (14.0, 29.0, 2.0, 11.5),
             ccrs.PlateCarree()),
        Area('cg', 'CONGO', (10.0, 20.0, -6.0, 5.0),
             ccrs.PlateCarree()),
        Area('ch', 'CHILE', (-80.0, -66.0, -56.0, -15.0),
             ccrs.SouthPolarStereo(central_longitude=-73)),
        Area('ci', 'CHINA', (85.0, 145.0, 14.0, 48.5),
             ccrs.NorthPolarStereo(central_longitude=110)),
        Area('cm', 'CAMEROON', (7.5, 17.1, 1.0, 14.0),
             ccrs.PlateCarree()),
        Area('colm', 'COLUMBIA', (-81.0, -65.0, -5.0, 14.0),
             ccrs.PlateCarree()),
        Area('cr', 'CANARY ISLANDS', (-19.0, -13.0, 27.0, 30.0),
             ccrs.PlateCarree()),
        Area('cs', 'COSTA RICA', (-86.5, -81.5, 8.2, 11.6),
             ccrs.PlateCarree()),
        Area('cu', 'CUBA', (-85.0, -74.0, 19.0, 24.0),
             ccrs.PlateCarree()),
        Area('cv', 'CAPE VERDE ISLAND', (-26.0, -22.0, 14.0, 18.0),
             ccrs.PlateCarree()),
        Area('cy', 'CYPRUS', (32.0, 35.0, 34.0, 36.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=34)),
        Area('cz', 'CZECH REPUBLIC', (8.9, 22.9, 47.4, 52.4),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=16)),
        Area('dj', 'DJIBOUTI', (41.5, 44.1, 10.5, 13.1),
             ccrs.PlateCarree()),
        Area('dl', 'GERMANY', (4.8, 16.8, 47.0, 55.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=11)),
        Area('dn', 'DENMARK', (8.0, 11.0, 54.0, 58.6),
             ccrs.NorthPolarStereo(central_longitude=10)),
        Area('do', 'DOMINICA', (-61.6, -61.2, 15.2, 15.8),
             ccrs.PlateCarree()),
        Area('dr', 'DOMINICAN REPUBLIC', (-72.2, -68.0, 17.5, 20.2),
             ccrs.PlateCarree()),
        Area('eg', 'EGYPT', (24.0, 37.0, 21.0, 33.0),
             ccrs.PlateCarree()),
        Area('eq', 'ECUADOR', (-85.0, -74.0, -7.0, 3.0),
             ccrs.PlateCarree()),
        Area('er', 'UAE', (50.0, 57.0, 22.0, 26.6),
             ccrs.PlateCarree()),
        Area('es', 'EL SALVADOR', (-90.3, -87.5, 13.0, 14.6),
             ccrs.PlateCarree()),
        Area('et', 'ETHIOPIA', (33.0, 49.0, 2.0, 19.0),
             ccrs.PlateCarree()),
        Area('fa', 'FAEROES', (-8.0, -6.0, 61.0, 63.0),
             ccrs.NorthPolarStereo(central_longitude=-7)),
        Area('fg', 'FRENCH GUIANA', (-55.0, -49.0, 1.0, 7.0),
             ccrs.PlateCarree()),
        Area('fi', 'FINLAND', (20.9, 35.1, 59.0, 70.6),
             ccrs.NorthPolarStereo(central_longitude=28)),
        Area('fj', 'FIJI ISLANDS', (176.0, 181.0, 16.0, 19.0),
             ccrs.PlateCarree(central_longitude=-180)),
        Area('fk', 'FALKLAND ISLANDS', (-61.3, -57.5, -53.0, -51.0),
             ccrs.SouthPolarStereo(central_longitude=-57)),
        Area('fn', 'NIGER', (0.0, 17.0, 11.0, 24.0),
             ccrs.PlateCarree()),
        Area('fr', 'FRANCE', (-5.0, 11.0, 41.0, 51.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=3)),
        Area('gb', 'GAMBIA', (-17.1, -13.5, 13.0, 14.6),
             ccrs.PlateCarree()),
        Area('gc', 'CAYMAN ISLANDS', (-82.8, -77.6, 17.9, 21.1),
             ccrs.PlateCarree()),
        Area('gh', 'GHANA', (-4.5, 1.5, 4.0, 12.0),
             ccrs.PlateCarree()),
        Area('gi', 'GIBRALTAR', (-8.0, -4.0, 35.0, 38.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-6)),
        Area('gl', 'GREENLAND', (-60, -20, 58.3, 85),
             ccrs.NorthPolarStereo(central_longitude=-45)),
        Area('glp', 'GUADALOUPE', (-64.2, -59.8, 14.8, 19.2),
             ccrs.PlateCarree()),
        Area('gm', 'GUAM', (144.5, 145.1, 13.0, 14.0),
             ccrs.PlateCarree()),
        Area('gn', 'GUINEA', (2.0, 16.0, 3.5, 15.5),
             ccrs.PlateCarree()),
        Area('go', 'GABON', (8.0, 14.5, -4.6, 3.0),
             ccrs.PlateCarree()),
        Area('gr', 'GREECE', (20.0, 27.6, 34.0, 42.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=24)),
        Area('gu', 'GUATEMALA', (-95.6, -85.0, 10.5, 21.1),
             ccrs.PlateCarree()),
        Area('gw', 'GUINEA-BISSAU', (-17.5, -13.5, 10.8, 12.8),
             ccrs.PlateCarree()),
        Area('gy', 'GUYANA', (-62.0, -55.0, 0.0, 10.0),
             ccrs.PlateCarree()),
        Area('ha', 'HAITI', (-75.0, -71.0, 18.0, 20.0),
             ccrs.PlateCarree()),
        Area('he', 'ST HELENA', (-6.1, -5.5, -16.3, -15.5),
             ccrs.PlateCarree()),
        Area('hk', 'HONG KONG', (113.5, 114.7, 22.0, 23.0),
             ccrs.PlateCarree()),
        Area('ho', 'HONDURAS', (-90.0, -83.0, 13.0, 16.6),
             ccrs.PlateCarree()),
        Area('hu', 'HUNGARY', (16.0, 23.0, 45.5, 49.1),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=20)),
        Area('ic', 'COMOROS', (43.0, 45.0, -13.2, -11.0),
             ccrs.PlateCarree()),
        Area('icel', 'ICELAND', (-24.1, -11.5, 63.0, 67.5),
             ccrs.NorthPolarStereo(central_longitude=-18)),
        Area('ie', 'IRELAND', (-11.1, -4.5, 50.0, 55.6),
             ccrs.NorthPolarStereo(central_longitude=-8)),
        Area('inda', 'INDIA', (67.0, 92.0, 4.2, 36.0),
             ccrs.PlateCarree()),
        Area('indo', 'INDONESIA', (95.0, 141.0, -8.0, 6.0),
             ccrs.PlateCarree()),
        Area('iq', 'IRAQ', (38.0, 50.0, 29.0, 38.0),
             ccrs.PlateCarree()),
        Area('ir', 'IRAN', (44.0, 65.0, 25.0, 40.0),
             ccrs.PlateCarree()),
        Area('is', 'ISRAEL', (34.0, 37.0, 29.0, 34.0),
             ccrs.PlateCarree()),
        Area('iv', 'IVORY COAST', (-9.0, -2.0, 4.0, 11.0),
             ccrs.PlateCarree()),
        Area('iw', 'WEST BANK', (34.8, 35.6, 31.2, 32.6),
             ccrs.PlateCarree()),
        Area('iy', 'ITALY', (6.6, 20.6, 35.6, 47.2),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=14)),
        Area('jd', 'JORDAN', (34.0, 39.6, 29.0, 33.6),
             ccrs.PlateCarree()),
        Area('jm', 'JAMAICA', (-80.0, -76.0, 16.0, 19.0),
             ccrs.PlateCarree()),
        Area('jp', 'JAPAN', (123.0, 155.0, 24.0, 47.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=139)),
        Area('ka', 'CAROLINE ISLANDS', (131.0, 155.0, 1.0, 9.6),
             ccrs.PlateCarree()),
        Area('kash', 'JAMMU & KASHMIR', (74.0, 78.0, 32.0, 35.0),
             ccrs.PlateCarree()),
        Area('kb', 'KIRIBATI', (172.0, 177.0, -3.0, 3.2),
             ccrs.PlateCarree()),
        Area('khm', 'CAMBODIA', (102.0, 108.0, 10.0, 15.0),
             ccrs.PlateCarree()),
        Area('ki', 'CHRISTMAS ISLAND', (105.2, 106.2, -11.0, -10.0),
             ccrs.PlateCarree()),
        Area('kn', 'KENYA', (32.5, 42.1, -6.0, 6.0),
             ccrs.PlateCarree()),
        Area('kna', 'ST KITTS & NEVIS', (-62.9, -62.4, 17.0, 17.5),
             ccrs.PlateCarree()),
        Area('ko', 'KOREA', (124.0, 131.5, 33.0, 43.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=128)),
        Area('ku', 'COOK ISLANDS', (-168.0, -155.0, -24.1, -6.1),
             ccrs.PlateCarree()),
        Area('kw', 'KUWAIT', (46.5, 48.5, 28.5, 30.5),
             ccrs.PlateCarree()),
        Area('laos', 'LAOS', (100.0, 108.0, 13.5, 23.1),
             ccrs.PlateCarree()),
        Area('lb', 'LEBANON', (34.5, 37.1, 33.0, 35.0),
             ccrs.PlateCarree()),
        Area('lc', 'ST LUCIA', (60.9, 61.3, 13.25, 14.45),
             ccrs.PlateCarree()),
        Area('li', 'LIBERIA', (-12.0, -7.0, 4.0, 9.0),
             ccrs.PlateCarree()),
        Area('ln', 'LINE ISLANDS', (-162.1, -154.9, -4.2, 6.0),
             ccrs.PlateCarree()),
        Area('ls', 'LESOTHO', (27.0, 29.6, -30.6, -28.0),
             ccrs.PlateCarree()),
        Area('lt', 'LIECHTENSTEIN', (9.3, 9.9, 47.0, 47.6),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=10)),
        Area('lux', 'LUXEMBOURG', (5.6, 6.6, 49.35, 50.25),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=6)),
        Area('ly', 'LIBYA', (8.0, 26.0, 19.0, 35.0),
             ccrs.PlateCarree()),
        Area('maar', 'ST MAARTEN', (-63.9, -62.3, 17.0, 18.6),
             ccrs.PlateCarree()),
        Area('made', 'MADEIRA', (-17.3, -16.5, 32.6, 33.0),
             ccrs.PlateCarree()),
        Area('mala', 'MALAYSIA', (100.0, 119.6, 1.0, 8.0),
             ccrs.PlateCarree()),
        Area('mali', 'MALI', (-12.5, 6.0, 8.5, 25.5),
             ccrs.PlateCarree()),
        Area('maur', 'MAURITIUS', (57.2, 57.8, -20.7, -19.9),
             ccrs.PlateCarree()),
        Area('maut', 'MAURITANIA', (-17.1, -4.5, 14.5, 28.1),
             ccrs.PlateCarree()),
        Area('mc', 'MOROCCO', (-13.0, -1.0, 25.0, 36.0),
             ccrs.PlateCarree()),
        Area('mg', 'MADAGASCAR', (43.0, 50.6, -25.6, -12.0),
             ccrs.PlateCarree()),
        Area('mh', 'MARSHALL ISLANDS', (160.0, 172.0, 4.5, 12.1),
             ccrs.PlateCarree()),
        Area('ml', 'MALTA', (14.3, 14.7, 35.8, 36.0),
             ccrs.PlateCarree()),
        Area('mmr', 'MYANMAR', (92.0, 102.0, 7.5, 28.5),
             ccrs.PlateCarree()),
        Area('mong', 'MONGOLIA', (87.5, 123.1, 38.5, 52.6),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=105)),
        Area('mr', 'MARTINIQUE', (-61.2, -60.8, 14.3, 15.1),
             ccrs.PlateCarree()),
        Area('mu', 'MACAO', (113.0, 114.0, 22.0, 23.0),
             ccrs.PlateCarree()),
        Area('mv', 'MALDIVE ISLANDS', (70.1, 76.1, -6.0, 10.0),
             ccrs.PlateCarree()),
        Area('mw', 'MALAWI', (32.5, 36.1, -17.0, -9.0),
             ccrs.PlateCarree()),
        Area('mx', 'MEXICO', (-119.0, -83.0, 13.0, 34.0),
             ccrs.PlateCarree()),
        Area('my', 'MARIANNA ISLANDS', (142.5, 148.5, 9.0, 25.0),
             ccrs.PlateCarree()),
        Area('mz', 'MOZAMBIQUE', (29.0, 41.0, -26.5, -9.5),
             ccrs.PlateCarree()),
        Area('nama', 'NAMIBIA', (11.0, 25.0, -29.5, -16.5),
             ccrs.PlateCarree()),
        Area('ncal', 'NEW CALEDONIA', (158.0, 172.0, -23.0, -18.0),
             ccrs.PlateCarree()),
        Area('ng', 'NEW GUINEA', (130.0, 152.0, -11.0, 0.0),
             ccrs.PlateCarree()),
        Area('ni', 'NIGERIA', (2.0, 14.6, 3.0, 14.0),
             ccrs.PlateCarree()),
        Area('nk', 'NICARAGUA', (-88.0, -83.0, 10.5, 15.1),
             ccrs.PlateCarree()),
        Area('nl', 'NETHERLANDS', (3.5, 7.5, 50.5, 54.1),
             ccrs.NorthPolarStereo(central_longitude=6)),
        Area('no', 'NORWAY', (3.0, 35.0, 57.0, 71.5),
             ccrs.NorthPolarStereo(central_longitude=19)),
        Area('np', 'NEPAL', (80.0, 89.0, 25.0, 31.0),
             ccrs.PlateCarree()),
        Area('nw', 'NAURU', (166.4, 167.4, -1.0, 0.0),
             ccrs.PlateCarree()),
        Area('nz', 'NEW ZEALAND', (165.0, 179.0, -48.0, -33.0),
             ccrs.SouthPolarStereo(central_longitude=172)),
        Area('om', 'OMAN', (52.0, 60.0, 16.0, 25.6),
             ccrs.PlateCarree()),
        Area('os', 'AUSTRIA', (9.0, 18.0, 46.0, 50.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=14)),
        Area('pf', 'FRENCH POLYNESIA', (-154.0, -134.0, -28.0, -8.0),
             ccrs.PlateCarree()),
        Area('ph', 'PHILIPINES', (116.0, 127.0, 4.0, 21.0),
             ccrs.PlateCarree()),
        Area('pi', 'PHOENIX ISLANDS', (-177.5, -167.5, -9.0, 1.0),
             ccrs.PlateCarree()),
        Area('pk', 'PAKISTAN', (60.0, 78.0, 23.0, 37.0),
             ccrs.PlateCarree()),
        Area('pl', 'POLAND', (14.0, 25.0, 48.5, 55.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=20)),
        Area('pm', 'PANAMA', (-83.0, -77.0, 7.0, 10.0),
             ccrs.PlateCarree()),
        Area('po', 'PORTUGAL', (-10.0, -4.0, 36.5, 42.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-7)),
        Area('pr', 'PERU', (-82.0, -68.0, -20.0, 5.0),
             ccrs.PlateCarree()),
        Area('pt', 'PITCAIRN ISLANDS', (-130.6, -129.6, -25.56, -24.56),
             ccrs.PlateCarree()),
        Area('pu', 'PUERTO RICO', (-67.5, -65.5, 17.5, 18.5),
             ccrs.PlateCarree()),
        Area('py', 'PARAGUAY', (-65.0, -54.0, -32.0, -17.0),
             ccrs.PlateCarree()),
        Area('qg', 'EQUATORIAL GUINEA', (7.0, 12.0, -2.0, 3.0),
             ccrs.PlateCarree()),
        Area('qt', 'QATAR', (50.0, 52.0, 24.0, 27.0),
             ccrs.PlateCarree()),
        Area('ra', 'RUSSIA', (40.0, 190.0, 35.0, 70.0),
             ccrs.LambertConformal(standard_parallels=[40, 70], central_longitude=105)),
        Area('re', 'REUNION', (55.0, 56.0, -21.5, -20.5),
             ccrs.PlateCarree()),
        Area('riro', 'RIO DE ORO', (-18.0, -12.0, 17.5, 27.5),
             ccrs.PlateCarree()),
        Area('ro', 'ROMANIA', (19.0, 31.0, 42.5, 48.5),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=25)),
        Area('rw', 'RWANDA', (29.0, 31.0, -3.0, -1.0),
             ccrs.PlateCarree()),
        Area('saudi', 'SAUDI ARABIA', (4.6, 92.5, -13.2, 60.3),
             ccrs.PlateCarree()),
        Area('sb', 'SRI LANKA', (79.0, 83.0, 5.0, 10.0),
             ccrs.PlateCarree()),
        Area('seyc', 'SEYCHELLES ISLANDS', (55.0, 56.0, -5.0, -4.0),
             ccrs.PlateCarree()),
        Area('sg', 'SENEGAL', (-18.0, -10.0, 12.0, 17.0),
             ccrs.PlateCarree()),
        Area('si', 'SOMALIA', (39.5, 52.1, -4.5, 13.5),
             ccrs.PlateCarree()),
        Area('sk', 'SARAWAK', (109.5, 119.3, 1.0, 7.0),
             ccrs.PlateCarree()),
        Area('sl', 'SIERRA LEONE', (-13.6, -10.2, 6.9, 10.1),
             ccrs.PlateCarree()),
        Area('sm', 'SURINAME', (-59.0, -53.0, 1.0, 6.0),
             ccrs.PlateCarree()),
        Area('sn', 'SWEDEN', (10.0, 25.0, 55.0, 69.6),
             ccrs.NorthPolarStereo(central_longitude=18)),
        Area('so', 'SOLOMON ISLANDS', (156.0, 167.0, -12.0, -6.0),
             ccrs.PlateCarree()),
        Area('sp', 'SPAIN', (-10.0, 6.0, 35.0, 44.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-2)),
        Area('sr', 'SINGAPORE', (103.0, 105.0, 1.0, 2.0),
             ccrs.PlateCarree()),
        Area('su', 'SUDAN', (21.5, 38.5, 3.5, 23.5),
             ccrs.PlateCarree()),
        Area('sv', 'SWAZILAND', (30.5, 33.1, -27.5, -25.3),
             ccrs.PlateCarree()),
        Area('sw', 'SWITZERLAND', (5.9, 10.5, 45.8, 48.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=8)),
        Area('sy', 'SYRIA', (35.0, 42.6, 32.0, 37.6),
             ccrs.PlateCarree()),
        Area('tanz', 'TANZANIA', (29.0, 40.6, -13.0, 0.0),
             ccrs.PlateCarree()),
        Area('td', 'TRINIDAD & TOBAGO', (-62.1, -60.5, 10.0, 11.6),
             ccrs.PlateCarree()),
        Area('tg', 'TOGO', (-0.5, 2.5, 5.0, 12.0),
             ccrs.PlateCarree()),
        Area('th', 'THAILAND', (97.0, 106.0, 5.0, 21.0),
             ccrs.PlateCarree()),
        Area('ti', 'TURKS ISLANDS', (-71.6, -70.6, 21.0, 22.0),
             ccrs.PlateCarree()),
        Area('tk', 'TOKELAU ISLANDS', (-173.0, -171.0, -11.5, -7.5),
             ccrs.PlateCarree()),
        Area('to', 'TONGA', (-178.5, -170.5, -22.0, -15.0),
             ccrs.PlateCarree()),
        Area('tp', 'SAO TOME-PRINCIPE', (6.0, 7.6, 0.0, 2.0),
             ccrs.PlateCarree()),
        Area('ts', 'TUNISIA', (7.0, 13.0, 30.0, 38.0),
             ccrs.PlateCarree()),
        Area('tu', 'TURKEY', (25.0, 48.0, 34.1, 42.1),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=37)),
        Area('tv', 'TUVALU', (176.0, 180.0, -11.0, -5.0),
             ccrs.PlateCarree()),
        Area('tw', 'TAIWAN', (120.0, 122.0, 21.9, 25.3),
             ccrs.PlateCarree()),
        Area('ug', 'UGANDA', (29.0, 35.0, -3.5, 5.5),
             ccrs.PlateCarree()),
        Area('uk', 'UNITED KINGDOM', (-11.0, 5.0, 49.0, 60.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=-3)),
        Area('ur', 'UKRAINE', (20.0, 41.0, 45.0, 52.0),
             ccrs.NorthPolarStereo(central_longitude=20)),
        Area('uy', 'URUGUAY', (-60.0, -52.0, -35.5, -29.5),
             ccrs.SouthPolarStereo(central_longitude=-56)),
        Area('vanu', 'VANUATU', (167.0, 170.0, -21.0, -13.0),
             ccrs.PlateCarree()),
        Area('vi', 'VIRGIN ISLANDS', (-65.5, -64.0, 16.6, 19.6),
             ccrs.PlateCarree()),
        Area('vk', 'SLOVAKIA', (13.8, 25.8, 46.75, 50.75),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=20)),
        Area('vn', 'VENEZUELA', (-75.0, -60.0, -2.0, 14.0),
             ccrs.PlateCarree()),
        Area('vs', 'VIETNAM', (102.0, 110.0, 8.0, 24.0),
             ccrs.PlateCarree()),
        Area('wk', 'WAKE ISLAND', (166.1, 167.1, 18.8, 19.8),
             ccrs.PlateCarree()),
        Area('ye', 'YEMEN', (42.5, 54.1, 12.5, 19.1),
             ccrs.PlateCarree()),
        Area('yg', 'YUGOSLAVIA', (13.5, 24.6, 40.0, 47.0),
             ccrs.LambertConformal(standard_parallels=[30, 60], central_longitude=19)),
        Area('za', 'SOUTH AFRICA', (16.0, 34.0, -36.0, -22.0),
             ccrs.PlateCarree()),
        Area('zb', 'ZAMBIA', (21.0, 35.0, -20.0, -7.0),
             ccrs.PlateCarree()),
        Area('zm', 'WESTERN SAMOA', (170.5, 173.5, -15.0, -13.0),
             ccrs.PlateCarree()),
        Area('zr', 'ZAIRE', (12.0, 31.6, -14.0, 6.0),
             ccrs.PlateCarree()),
        Area('zw', 'ZIMBABWE', (25.0, 34.0, -22.9, -15.5),
             ccrs.PlateCarree())
    ]

    named_areas = {data.name: data for data in area_data}
except ImportError:
    # If no Cartopy is present, we just don't have plot areas
    pass

__all__ = ['named_areas']



# Copyright (c) 2020 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Test the `_tools` module."""

from metpy.io._tools import NamedStruct


def test_unpack():
    """Test unpacking a NamedStruct from bytes."""
    struct = NamedStruct([('field1', 'i'), ('field2', 'h')], '>')

    s = struct.unpack(b'\x00\x01\x00\x01\x00\x02')
    assert s.field1 == 65537
    assert s.field2 == 2


def test_pack():
    """Test packing a NamedStruct into bytes."""
    struct = NamedStruct([('field1', 'i'), ('field2', 'h')], '>')

    b = struct.pack(field1=8, field2=3)
    assert b == b'\x00\x00\x00\x08\x00\x03'



# Copyright (c) 2015,2016,2017 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Test the `gini` module."""

from datetime import datetime
import logging

import numpy as np
from numpy.testing import assert_almost_equal
import pytest
import xarray as xr

from metpy.cbook import get_test_data
from metpy.io import GiniFile
from metpy.io.gini import GiniProjection

logging.getLogger('metpy.io.gini').setLevel(logging.ERROR)

# Reference contents of the named tuples from each file
make_pdb = GiniFile.prod_desc_fmt.make_tuple
make_pdb2 = GiniFile.prod_desc2_fmt.make_tuple
raw_gini_info = [('WEST-CONUS_4km_WV_20151208_2200.gini',
                  make_pdb(source=1, creating_entity='GOES-15', sector_id='West CONUS',
                           channel='WV (6.5/6.7 micron)', num_records=1280, record_len=1100,
                           datetime=datetime(2015, 12, 8, 22, 0, 19),
                           projection=GiniProjection.lambert_conformal, nx=1100, ny=1280,
                           la1=12.19, lo1=-133.4588),
                  make_pdb2(scanning_mode=[False, False, False], lat_in=25.0, resolution=4,
                            compression=0, version=1, pdb_size=512, nav_cal=0),
                  GiniFile.lc_ps_fmt.make_tuple(reserved=0, lov=-95.0, dx=4.0635, dy=4.0635,
                                                proj_center=0)),
                 ('AK-REGIONAL_8km_3.9_20160408_1445.gini',
                  make_pdb(source=1, creating_entity='GOES-15', sector_id='Alaska Regional',
                           channel='IR (3.9 micron)', num_records=408, record_len=576,
                           datetime=datetime(2016, 4, 8, 14, 45, 20),
                           projection=GiniProjection.polar_stereographic, nx=576, ny=408,
                           la1=42.0846, lo1=-175.641),
                  make_pdb2(scanning_mode=[False, False, False], lat_in=0.0, resolution=8,
                            compression=0, version=1, pdb_size=512, nav_cal=0),
                  GiniFile.lc_ps_fmt.make_tuple(reserved=0, lov=210.0, dx=7.9375, dy=7.9375,
                                                proj_center=0)),
                 ('HI-REGIONAL_4km_3.9_20160616_1715.gini',
                  make_pdb(source=1, creating_entity='GOES-15', sector_id='Hawaii Regional',
                           channel='IR (3.9 micron)', num_records=520, record_len=560,
                           datetime=datetime(2016, 6, 16, 17, 15, 18),
                           projection=GiniProjection.mercator, nx=560, ny=520,
                           la1=9.343, lo1=-167.315),
                  make_pdb2(scanning_mode=[False, False, False], lat_in=20.0, resolution=4,
                            compression=0, version=1, pdb_size=512, nav_cal=0),
                  GiniFile.mercator_fmt.make_tuple(resolution=0, la2=28.0922, lo2=-145.878,
                                                   di=0, dj=0))
                 ]


@pytest.mark.parametrize('filename,pdb,pdb2,proj_info', raw_gini_info,
                         ids=['LCC', 'Stereographic', 'Mercator'])
def test_raw_gini(filename, pdb, pdb2, proj_info):
    """Test raw GINI parsing."""
    f = GiniFile(get_test_data(filename))
    assert f.prod_desc == pdb
    assert f.prod_desc2 == pdb2
    assert f.proj_info == proj_info
    assert f.data.shape == (pdb.num_records, pdb.record_len)


def test_gini_bad_size():
    """Test reading a GINI file that reports a bad header size."""
    f = GiniFile(get_test_data('NHEM-MULTICOMP_1km_IR_20151208_2100.gini'))
    pdb2 = f.prod_desc2
    assert pdb2.pdb_size == 512  # Catching bad size


# Reference information coming out of the dataset interface, coordinate calculations,
# inclusion of correct projection metadata, etc.
gini_dataset_info = [('WEST-CONUS_4km_WV_20151208_2200.gini',
                      (-4226066.37649, 239720.12351, -832700.70519, 4364515.79481), 'WV',
                      {'grid_mapping_name': 'lambert_conformal_conic',
                       'standard_parallel': 25.0, 'earth_radius': 6371200.,
                       'latitude_of_projection_origin': 25.0,
                       'longitude_of_central_meridian': -95.0}, (150, 150, 184),
                      datetime(2015, 12, 8, 22, 0, 19)),
                     ('AK-REGIONAL_8km_3.9_20160408_1445.gini',
                      (-2286001.13195, 2278061.36805, -4762503.5992, -1531941.0992), 'IR',
                      {'grid_mapping_name': 'polar_stereographic', 'standard_parallel': 60.0,
                       'earth_radius': 6371200., 'latitude_of_projection_origin': 90.,
                       'straight_vertical_longitude_from_pole': 210.0}, (150, 150, 143),
                      datetime(2016, 4, 8, 14, 45, 20)),
                     ('HI-REGIONAL_4km_3.9_20160616_1715.gini',
                      (0.0, 2236000.0, 980627.44738, 3056627.44738), 'IR',
                      {'grid_mapping_name': 'mercator', 'standard_parallel': 20.0,
                       'earth_radius': 6371200., 'latitude_of_projection_origin': 9.343,
                       'longitude_of_projection_origin': -167.315}, (150, 150, 111),
                      datetime(2016, 6, 16, 17, 15, 18))
                     ]


@pytest.mark.parametrize('filename,bounds,data_var,proj_attrs,image,dt', gini_dataset_info,
                         ids=['LCC', 'Stereographic', 'Mercator'])
def test_gini_xarray(filename, bounds, data_var, proj_attrs, image, dt):
    """Test that GINIFile can be passed to XArray as a datastore."""
    f = GiniFile(get_test_data(filename))
    ds = xr.open_dataset(f)

    # Check our calculated x and y arrays
    x0, x1, y0, y1 = bounds
    x = ds.variables['x']
    assert_almost_equal(x[0], x0, 4)
    assert_almost_equal(x[-1], x1, 4)

    # Because the actual data raster has the top row first, the maximum y value is y[0],
    # while the minimum y value is y[-1]
    y = ds.variables['y']
    assert_almost_equal(y[-1], y0, 4)
    assert_almost_equal(y[0], y1, 4)

    # Check the projection metadata
    proj_name = ds.variables[data_var].attrs['grid_mapping']
    proj_var = ds.variables[proj_name]
    for attr, val in proj_attrs.items():
        assert proj_var.attrs[attr] == val, 'Values mismatch for ' + attr

    # Check the lower left lon/lat corner
    assert_almost_equal(ds.variables['lon'][-1, 0], f.prod_desc.lo1, 4)
    assert_almost_equal(ds.variables['lat'][-1, 0], f.prod_desc.la1, 4)

    # Check a pixel of the image to make sure we're decoding correctly
    x_ind, y_ind, val = image
    assert val == ds.variables[data_var][x_ind, y_ind]

    # Check time decoding
    assert np.asarray(dt, dtype='datetime64[ms]') == ds.variables['time']


@pytest.mark.parametrize('filename,bounds,data_var,proj_attrs,image,dt', gini_dataset_info,
                         ids=['LCC', 'Stereographic', 'Mercator'])
@pytest.mark.parametrize('specify_engine', [True, False], ids=['engine', 'no engine'])
def test_gini_xarray_entrypoint(filename, bounds, data_var, proj_attrs, image, dt,
                                specify_engine):
    """Test that GINI files can be opened directly by xarray using the v2 API."""
    ds = xr.open_dataset(get_test_data(filename, as_file_obj=specify_engine),
                         **({'engine': 'gini'} if specify_engine else {}))

    # Check our calculated x and y arrays
    x0, x1, y0, y1 = bounds
    x = ds.variables['x']
    assert_almost_equal(x[0], x0, 4)
    assert_almost_equal(x[-1], x1, 4)

    # Because the actual data raster has the top row first, the maximum y value is y[0],
    # while the minimum y value is y[-1]
    y = ds.variables['y']
    assert_almost_equal(y[-1], y0, 4)
    assert_almost_equal(y[0], y1, 4)

    # Check the projection metadata
    proj_name = ds.variables[data_var].attrs['grid_mapping']
    proj_var = ds.variables[proj_name]
    for attr, val in proj_attrs.items():
        assert proj_var.attrs[attr] == val, 'Values mismatch for ' + attr

    # Check a pixel of the image to make sure we're decoding correctly
    x_ind, y_ind, val = image
    assert val == ds.variables[data_var][x_ind, y_ind]

    # Check time decoding
    assert 'time' in ds.coords
    assert np.asarray(dt, dtype='datetime64[ms]') == ds.variables['time']


def test_gini_mercator_upper_corner():
    """Test that the upper corner of the Mercator coordinates is correct."""
    f = GiniFile(get_test_data('HI-REGIONAL_4km_3.9_20160616_1715.gini'))
    ds = xr.open_dataset(f)
    lat = ds.variables['lat']
    lon = ds.variables['lon']

    # 2nd corner lat/lon are at the "upper right" corner of the pixel, so need to add one
    # more grid point
    assert_almost_equal(lon[0, -1] + (lon[0, -1] - lon[0, -2]), f.proj_info.lo2, 4)
    assert_almost_equal(lat[0, -1] + (lat[0, -1] - lat[1, -1]), f.proj_info.la2, 4)


def test_gini_str():
    """Test the str representation of GiniFile."""
    f = GiniFile(get_test_data('WEST-CONUS_4km_WV_20151208_2200.gini'))
    truth = ('GiniFile: GOES-15 West CONUS WV (6.5/6.7 micron)\n'
             '\tTime: 2015-12-08 22:00:19\n\tSize: 1280x1100\n'
             '\tProjection: lambert_conformal\n'
             '\tLower Left Corner (Lon, Lat): (-133.4588, 12.19)\n\tResolution: 4km')
    assert str(f) == truth


def test_gini_pathlib():
    """Test that GiniFile works with `pathlib.Path` instances."""
    from pathlib import Path
    src = Path(get_test_data('WEST-CONUS_4km_WV_20151208_2200.gini', as_file_obj=False))
    f = GiniFile(src)
    assert f.prod_desc.sector_id == 'West CONUS'


def test_unidata_composite():
    """Test reading radar composites in GINI format made by Unidata."""
    f = GiniFile(get_test_data('Level3_Composite_dhr_1km_20180309_2225.gini'))

    # Check the time stamp
    assert datetime(2018, 3, 9, 22, 25) == f.prod_desc.datetime

    # Check data value
    assert f.data[2160, 2130] == 66


def test_percent_normal():
    """Test reading PCT products properly."""
    f = GiniFile(get_test_data('PR-NATIONAL_1km_PCT_20200320_0446.gini'))

    assert f.prod_desc.channel == 'Percent Normal TPW'



# Copyright (c) 2015,2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tools for reading, calculating, and plotting with weather data."""

# What do we want to pull into the top-level namespace?
import warnings

# Must occur before below imports
warnings.filterwarnings('ignore', 'numpy.dtype size changed')

from ._version import get_version  # noqa: E402
from .xarray import *  # noqa: F401, F403, E402

__version__ = get_version()
del get_version



# Copyright (c) 2015,2016,2017,2021 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Test the `gempak` module."""

from datetime import datetime
import logging

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal, assert_equal
import pandas as pd
import pytest

from metpy.cbook import get_test_data
from metpy.io.gempak import GempakGrid, GempakSounding, GempakSurface

logging.getLogger('metpy.io.gempak').setLevel(logging.ERROR)


@pytest.mark.parametrize('order', ['little', 'big'])
def test_byte_swap(order):
    """"Test byte swapping."""
    g = get_test_data(f'gem_{order}_endian.grd')

    grid = GempakGrid(g).gdxarray()[0].squeeze()

    reference = np.ones((113, 151), dtype='int32')

    assert_equal(grid, reference)


@pytest.mark.parametrize('grid_name', ['none', 'diff', 'dec', 'grib'])
def test_grid_loading(grid_name):
    """Test reading grids with different packing."""
    grid = GempakGrid(get_test_data(f'gem_packing_{grid_name}.grd')).gdxarray(
        parameter='TMPK',
        level=850
    )
    gio = grid[0].values.squeeze()

    gempak = np.load(get_test_data(f'gem_packing_{grid_name}.npz',
                                   as_file_obj=False))['values']

    assert_allclose(gio, gempak, rtol=1e-6, atol=0)


def test_merged_sounding():
    """Test loading a merged sounding.

    These are most often from models.
    """
    gso = GempakSounding(get_test_data('gem_model_mrg.snd')).snxarray(
        station_id='KMSN'
    )
    gpres = gso[0].pressure.values
    gtemp = gso[0].tmpc.values.squeeze()
    gdwpt = gso[0].dwpc.values.squeeze()
    gdrct = gso[0].drct.values.squeeze()
    gsped = gso[0].sped.values.squeeze()
    ghght = gso[0].hght.values.squeeze()
    gomeg = gso[0].omeg.values.squeeze()
    gcwtr = gso[0].cwtr.values.squeeze()
    gdtcp = gso[0].dtcp.values.squeeze()
    gdtgp = gso[0].dtgp.values.squeeze()
    gdtsw = gso[0].dtsw.values.squeeze()
    gdtlw = gso[0].dtlw.values.squeeze()
    gcfrl = gso[0].cfrl.values.squeeze()
    gtkel = gso[0].tkel.values.squeeze()
    gimxr = gso[0].imxr.values.squeeze()
    gdtar = gso[0].dtar.values.squeeze()

    gempak = pd.read_csv(get_test_data('gem_model_mrg.csv', as_file_obj=False),
                         na_values=-9999)
    dpres = gempak.PRES.values
    dtemp = gempak.TMPC.values
    ddwpt = gempak.DWPC.values
    ddrct = gempak.DRCT.values
    dsped = gempak.SPED.values
    dhght = gempak.HGHT.values
    domeg = gempak.OMEG.values
    dcwtr = gempak.CWTR.values
    ddtcp = gempak.DTCP.values
    ddtgp = gempak.DTGP.values
    ddtsw = gempak.DTSW.values
    ddtlw = gempak.DTLW.values
    dcfrl = gempak.CFRL.values
    dtkel = gempak.TKEL.values
    dimxr = gempak.IMXR.values
    ddtar = gempak.DTAR.values

    np.testing.assert_allclose(gpres, dpres, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gtemp, dtemp, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdwpt, ddwpt, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdrct, ddrct, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gsped, dsped, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(ghght, dhght, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gomeg, domeg, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gcwtr, dcwtr, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdtcp, ddtcp, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdtgp, ddtgp, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdtsw, ddtsw, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdtlw, ddtlw, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gcfrl, dcfrl, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gtkel, dtkel, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gimxr, dimxr, rtol=1e-10, atol=1e-2)
    np.testing.assert_allclose(gdtar, ddtar, rtol=1e-10, atol=1e-2)


def test_merged_sounding_no_packing():
    """Test loading a merged sounding without data packing."""
    gso = GempakSounding(get_test_data('gem_merged_nopack.snd')).snxarray(
        station_id='OUN')

    gpres = gso[0].pressure.values
    gtemp = gso[0].temp.values.squeeze()
    gdwpt = gso[0].dwpt.values.squeeze()
    gdrct = gso[0].drct.values.squeeze()
    gsped = gso[0].sped.values.squeeze()
    ghght = gso[0].hght.values.squeeze()

    gempak = pd.read_csv(get_test_data('gem_merged_nopack.csv', as_file_obj=False),
                         na_values=-9999)
    dpres = gempak.PRES.values
    dtemp = gempak.TEMP.values
    ddwpt = gempak.DWPT.values
    ddrct = gempak.DRCT.values
    dsped = gempak.SPED.values
    dhght = gempak.HGHT.values

    assert_allclose(gpres, dpres, rtol=1e-10, atol=1e-2)
    assert_allclose(gtemp, dtemp, rtol=1e-10, atol=1e-2)
    assert_allclose(gdwpt, ddwpt, rtol=1e-10, atol=1e-2)
    assert_allclose(gdrct, ddrct, rtol=1e-10, atol=1e-2)
    assert_allclose(gsped, dsped, rtol=1e-10, atol=1e-2)
    assert_allclose(ghght, dhght, rtol=1e-10, atol=1e-1)


@pytest.mark.parametrize('gem,gio,station', [
    ('gem_sigw_hght_unmrg.csv', 'gem_sigw_hght_unmrg.snd', 'TOP'),
    ('gem_sigw_pres_unmrg.csv', 'gem_sigw_pres_unmrg.snd', 'WAML')
])
def test_unmerged_sounding(gem, gio, station):
    """Test loading an unmerged sounding.

    PPBB and PPDD groups will be in height coordinates.
    """
    gso = GempakSounding(get_test_data(f'{gio}')).snxarray(
        station_id=f'{station}'
    )
    gpres = gso[0].pressure.values
    gtemp = gso[0].temp.values.squeeze()
    gdwpt = gso[0].dwpt.values.squeeze()
    gdrct = gso[0].drct.values.squeeze()
    gsped = gso[0].sped.values.squeeze()
    ghght = gso[0].hght.values.squeeze()

    gempak = pd.read_csv(get_test_data(f'{gem}', as_file_obj=False), na_values=-9999)
    dpres = gempak.PRES.values
    dtemp = gempak.TEMP.values
    ddwpt = gempak.DWPT.values
    ddrct = gempak.DRCT.values
    dsped = gempak.SPED.values
    dhght = gempak.HGHT.values

    assert_allclose(gpres, dpres, rtol=1e-10, atol=1e-2)
    assert_allclose(gtemp, dtemp, rtol=1e-10, atol=1e-2)
    assert_allclose(gdwpt, ddwpt, rtol=1e-10, atol=1e-2)
    assert_allclose(gdrct, ddrct, rtol=1e-10, atol=1e-2)
    assert_allclose(gsped, dsped, rtol=1e-10, atol=1e-2)
    assert_allclose(ghght, dhght, rtol=1e-10, atol=1e-1)


def test_unmerged_sigw_pressure_sounding():
    """Test loading an unmerged sounding.

    PPBB and PPDD groups will be in pressure coordinates and there will
    be MAN level below the surface.
    """
    gso = GempakSounding(get_test_data('gem_sigw_pres_unmrg_man_bgl.snd')).snxarray()
    gpres = gso[0].pressure.values
    gtemp = gso[0].temp.values.squeeze()
    gdwpt = gso[0].dwpt.values.squeeze()
    gdrct = gso[0].drct.values.squeeze()
    gsped = gso[0].sped.values.squeeze()
    ghght = gso[0].hght.values.squeeze()

    gempak = pd.read_csv(get_test_data('gem_sigw_pres_unmrg_man_bgl.csv', as_file_obj=False),
                         na_values=-9999)
    dpres = gempak.PRES.values
    dtemp = gempak.TEMP.values
    ddwpt = gempak.DWPT.values
    ddrct = gempak.DRCT.values
    dsped = gempak.SPED.values
    dhght = gempak.HGHT.values

    assert_allclose(gpres, dpres, rtol=1e-10, atol=1e-2)
    assert_allclose(gtemp, dtemp, rtol=1e-10, atol=1e-2)
    assert_allclose(gdwpt, ddwpt, rtol=1e-10, atol=1e-2)
    assert_allclose(gdrct, ddrct, rtol=1e-10, atol=1e-2)
    assert_allclose(gsped, dsped, rtol=1e-10, atol=1e-2)
    assert_allclose(ghght, dhght, rtol=1e-10, atol=1e-1)


def test_climate_surface():
    """Test to read a cliamte surface file."""
    gsf = GempakSurface(get_test_data('gem_climate.sfc'))
    gstns = gsf.sfjson()

    gempak = pd.read_csv(get_test_data('gem_climate.csv', as_file_obj=False))
    gempak['YYMMDD/HHMM'] = pd.to_datetime(gempak['YYMMDD/HHMM'], format='%y%m%d/%H%M')
    gempak = gempak.set_index(['STN', 'YYMMDD/HHMM'])

    for stn in gstns:
        idx_key = (stn['properties']['station_id'],
                   stn['properties']['date_time'])
        gemsfc = gempak.loc[idx_key, :]

        for param, val in stn['values'].items():
            assert val == pytest.approx(gemsfc[param.upper()])


def test_standard_surface():
    """Test to read a standard surface file."""
    skip = ['text', 'spcl']

    gsf = GempakSurface(get_test_data('gem_std.sfc'))
    gstns = gsf.sfjson()

    gempak = pd.read_csv(get_test_data('gem_std.csv', as_file_obj=False))
    gempak['YYMMDD/HHMM'] = pd.to_datetime(gempak['YYMMDD/HHMM'], format='%y%m%d/%H%M')
    gempak = gempak.set_index(['STN', 'YYMMDD/HHMM'])

    for stn in gstns:
        idx_key = (stn['properties']['station_id'],
                   stn['properties']['date_time'])
        gemsfc = gempak.loc[idx_key, :]

        for param, val in stn['values'].items():
            if param not in skip:
                assert val == pytest.approx(gemsfc[param.upper()])


def test_ship_surface():
    """Test to read a ship surface file."""
    skip = ['text', 'spcl']

    gsf = GempakSurface(get_test_data('gem_ship.sfc'))

    gempak = pd.read_csv(get_test_data('gem_ship.csv', as_file_obj=False))
    gempak['YYMMDD/HHMM'] = pd.to_datetime(gempak['YYMMDD/HHMM'], format='%y%m%d/%H%M')
    gempak = gempak.set_index(['STN', 'YYMMDD/HHMM'])
    gempak.sort_index(inplace=True)

    uidx = gempak.index.unique()

    for stn, dt in uidx:
        ugem = gempak.loc[(stn, dt), ]
        gstns = gsf.sfjson(station_id=stn, date_time=dt)

        assert len(ugem) == len(gstns)

        params = gempak.columns
        for param in params:
            if param not in skip:
                decoded_vals = [d['values'][param.lower()] for d in gstns]
                actual_vals = ugem.loc[:, param].values
                assert_allclose(decoded_vals, actual_vals)


@pytest.mark.parametrize('proj_type', ['conical', 'cylindrical', 'azimuthal'])
def test_coordinates_creation(proj_type):
    """Test projections and coordinates."""
    grid = GempakGrid(get_test_data(f'gem_{proj_type}.grd'))
    decode_lat = grid.lat
    decode_lon = grid.lon

    gempak = np.load(get_test_data(f'gem_{proj_type}.npz', as_file_obj=False))
    true_lat = gempak['lat']
    true_lon = gempak['lon']

    assert_allclose(decode_lat, true_lat, rtol=1e-6, atol=1e-2)
    assert_allclose(decode_lon, true_lon, rtol=1e-6, atol=1e-2)


@pytest.mark.parametrize('proj_type, proj_attrs', [
    ('conical', {
        'grid_mapping_name': 'lambert_conformal_conic',
        'standard_parallel': (25.0, 25.0),
        'latitude_of_projection_origin': 0.0,
        'longitude_of_central_meridian': -95.0,
        'semi_major_axis': 6371200.,
        'semi_minor_axis': 6371200.,
    }),
    ('azimuthal', {
        'grid_mapping_name': 'polar_stereographic',
        'latitude_of_projection_origin': 90.0,
        'straight_vertical_longitude_from_pole': -105.0,
        'scale_factor_at_projection_origin': 1.0,
        'semi_major_axis': 6371200.,
        'semi_minor_axis': 6371200.,
    })
])
def test_metpy_crs_creation(proj_type, proj_attrs):
    """Test grid mapping metadata."""
    grid = GempakGrid(get_test_data(f'gem_{proj_type}.grd'))
    arr = grid.gdxarray()[0]
    metpy_crs = arr.metpy.crs
    for k, v in proj_attrs.items():
        assert metpy_crs[k] == v
    x_unit = arr['x'].units
    y_unit = arr['y'].units
    assert x_unit == 'meters'
    assert y_unit == 'meters'


def test_date_parsing():
    """Test parsing of dates with leading zeroes."""
    sfc_data = GempakSurface(get_test_data('sfc_obs.gem'))
    dat = sfc_data.sfinfo()[0].DATTIM
    assert dat == datetime(2000, 1, 2)


@pytest.mark.parametrize('access_type', ['STID', 'STNM'])
def test_surface_access(access_type):
    """Test for proper surface retrieval with multi-parameter filter."""
    g = get_test_data('gem_surface_with_text.sfc')
    gsf = GempakSurface(g)

    if access_type == 'STID':
        gsf.sfjson(station_id='MSN', country='US', state='WI',
                   date_time='202109070000')
    elif access_type == 'STNM':
        gsf.sfjson(station_number=726410, country='US', state='WI',
                   date_time='202109070000')


@pytest.mark.parametrize('text_type,date_time', [
    ('text', '202109070000'), ('spcl', '202109071600')
])
def test_surface_text(text_type, date_time):
    """Test text decoding of surface hourly and special observations."""
    g = get_test_data('gem_surface_with_text.sfc')
    gsf = GempakSurface(g)
    text = gsf.nearest_time(date_time, station_id='MSN')[0]['values'][text_type]

    gempak = pd.read_csv(get_test_data('gem_surface_with_text.csv', as_file_obj=False))
    gem_text = gempak.loc[:, text_type.upper()][0]

    assert text == gem_text


@pytest.mark.parametrize('access_type', ['STID', 'STNM'])
def test_sounding_access(access_type):
    """Test for proper sounding retrieval with multi-parameter filter."""
    g = get_test_data('gem_merged_nopack.snd')
    gso = GempakSounding(g)

    if access_type == 'STID':
        gso.snxarray(station_id='OUN', country='US', state='OK',
                     date_time='202101200000')
    elif access_type == 'STNM':
        gso.snxarray(station_number=72357, country='US', state='OK',
                     date_time='202101200000')


@pytest.mark.parametrize('text_type', ['txta', 'txtb', 'txtc', 'txpb'])
def test_sounding_text(text_type):
    """Test for proper decoding of coded message text."""
    g = get_test_data('gem_unmerged_with_text.snd')
    gso = GempakSounding(g).snxarray(station_id='OUN')[0]

    gempak = pd.read_csv(get_test_data('gem_unmerged_with_text.csv', as_file_obj=False))

    text = gso.attrs['WMO_CODES'][text_type]
    gem_text = gempak.loc[:, text_type.upper()][0]

    assert text == gem_text


def test_special_surface_observation():
    """Test special surface observation conversion."""
    sfc = get_test_data('gem_surface_with_text.sfc')

    gsf = GempakSurface(sfc)
    stn = gsf.nearest_time('202109071601',
                           station_id='MSN',
                           include_special=True)[0]['values']

    assert_almost_equal(stn['pmsl'], 1003.81, 2)
    assert stn['alti'] == 29.66
    assert stn['tmpc'] == 22
    assert stn['dwpc'] == 18
    assert stn['sknt'] == 9
    assert stn['drct'] == 230
    assert stn['gust'] == 18
    assert stn['wnum'] == 77
    assert stn['chc1'] == 2703
    assert stn['chc2'] == 8004
    assert stn['chc3'] == -9999
    assert stn['vsby'] == 2


def test_multi_level_multi_time_access():
    """Test accessing data with multiple levels and times."""
    g = get_test_data('gem_multilevel_multidate.grd')

    grid = GempakGrid(g)

    grid.gdxarray(
        parameter='STPC',
        date_time='202403040000',
        coordinate='HGHT',
        level=0,
        date_time2='202403050000',
        level2=1
    )


def test_multi_time_grid():
    """Test files with multiple times on a single grid."""
    g = get_test_data('gem_multi_time.grd')

    grid = GempakGrid(g)
    grid_info = grid.gdinfo()[0]
    dattim1 = grid_info.DATTIM1
    dattim2 = grid_info.DATTIM2

    assert dattim1 == datetime(1991, 8, 19, 0, 0)
    assert dattim2 == datetime(1991, 8, 20, 0, 0)



#!/usr/bin/env python3
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

from datetime import datetime, timezone
import inspect
import os
from pathlib import Path
import re
import sys

import metpy

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use resolve() to make it absolute, like shown here.
cwd = Path.cwd().resolve()
sys.path.insert(0, str(cwd))
sys.path.insert(0, str(cwd.parent.parent))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '2.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.linkcode',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx_design',
    'sphinx_gallery.gen_gallery',
    'matplotlib.sphinxext.plot_directive',
    'myst_parser',
    'make_areas'
]

sphinx_gallery_conf = {
    'doc_module': ('metpy',),
    'reference_url': {
        'metpy': None,
    },
    'examples_dirs': [str(cwd.parent / 'examples'), str(cwd.parent / 'tutorials')],
    'gallery_dirs': ['examples', 'tutorials'],
    'filename_pattern': r'\.py',
    'backreferences_dir': str(Path('api') / 'generated'),
    'default_thumb_file': str(Path('_static') / 'metpy_150x150_white_bg.png'),
    'abort_on_example_error': True,
    'reset_modules': [lambda conf, fname: sys.modules.pop('pint', None)]
}

# By default, only generate all the areas when running on a release CI job
metpy_generate_all_areas = (not os.environ.get('GITHUB_REF', '').startswith('refs/pull')
                            and sys.version_info < (3, 12))

# Turn off code and image links for embedded mpl plots
plot_html_show_source_link = False
plot_html_show_formats = False
plot_formats = ['png']
plot_rcparams = {'savefig.bbox': 'tight'}

# Make MyST generate anchors for headings
myst_heading_anchors = 2

# Set up mapping for other projects' docs
intersphinx_mapping = {
                       'cartopy': ('https://scitools.org.uk/cartopy/docs/latest/', None),
                       'matplotlib': ('https://matplotlib.org/stable/', None),
                       'numpy': ('https://numpy.org/doc/stable/', None),
                       'pandas': ('https://pandas.pydata.org/docs/', None),
                       'pint': ('https://pint.readthedocs.io/en/stable/', None),
                       'pyproj': ('https://pyproj4.github.io/pyproj/stable/', None),
                       'python': ('https://docs.python.org/3/', None),
                       'scipy': ('https://docs.scipy.org/doc/scipy/', None),
                       'xarray': ('https://docs.xarray.dev/en/stable/', None)
                       }

nitpicky = True
nitpick_ignore = [
    ('py:class', 'M'), ('py:class', 'N'), ('py:class', 'P'), ('py:class', '2'),
    ('py:class', 'optional'), ('py:class', 'array-like'), ('py:class', 'file-like object'),
    # For traitlets docstrings
    ('py:class', 'All'), ('py:class', 't.Any'), ('py:class', 't.Iterable'),
    # Next two are from Python dict docstring that we inherit
    ('py:class', 'a shallow copy of D'),
    ('py:class', 'v, remove specified key and return the corresponding value.')
]

nitpick_ignore_regex = [
    ('py:class', r'default:.*'),  # For some traitlets docstrings
    ('py:class', r'.*Sentinel'),
    ('py:class', r'.*Bunch'),
    ('py:class', r'.*[cC]allable'),
    ('py:class', r'.*EventHandler'),
    ('py:class', r'.*TraitType'),
    ('py:class', r'.*object providing a view on.*'),  # Python dict docstring
    ('py:class', r'None.  .*'),  # Python dict docstring
    ('py:class', r'.*D\[k\].*'),  # Python dict docstring
]

show_warning_types = True

# Eliminates unnecessary warning about sphinx-gallery rebuild
suppress_warnings = ['config.cache']

# Tweak how docs are formatted
napoleon_use_rtype = False

# Control main class documentation
autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = ['.rst', '.md']

# Controlling automatically generating summary tables in the docs
autosummary_generate = True
autosummary_ignore_module_all = False

# The encoding of source files.
# source_encoding = 'utf-8-sig'
cur_date = datetime.now(timezone.utc)

# The main toctree document.
master_doc = 'index'

# General information about the project.
project = 'MetPy'

# noinspection PyShadowingBuiltins
copyright = (
    f'2008\u2013{cur_date:%Y}, MetPy Developers. '
    'Development is supported by Unidata and the National Science Foundation'
)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
verinfo = metpy.__version__
parsed_version = re.search(r'(?P<full>(?P<base>\d+\.\d+)\.?\w*)', verinfo).groupdict()
# The short X.Y version.
version = parsed_version['base']
if '+' in verinfo:
    version += 'dev'
# The full version, including alpha/beta/rc tags.
release = parsed_version['full']

rst_prolog = f'''
.. |cite_version| replace:: {release}
.. |cite_year| replace:: {cur_date:%Y}
.. |access_date| replace:: {cur_date:%d %B %Y}
'''

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'autolink'

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'pydata_sphinx_theme'

# Use the version set in CI as necessary, which allows building "release" docs on a
# maintenance branch--strip leading 'v' since our json file doesn't have the v on the 'version'
doc_version = os.environ.get('DOC_VERSION', 'dev' if 'dev' in version else version).lstrip('v')
html_theme_options = {
    'external_links': [
        {'name': 'Release Notes', 'url': 'https://github.com/Unidata/MetPy/releases'},
    ],
    'icon_links': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/Unidata/MetPy',
            'icon': 'fa-brands fa-github',
            'type': 'fontawesome',
        },
        {
            'name': 'Twitter',
            'url': 'https://twitter.com/MetPy',
            'icon': 'fa-brands fa-twitter',
            'type': 'fontawesome',
        },
        {
            'name': 'Calendar',
            'url': 'https://calendar.google.com/calendar/embed?src=c_596cc34cd7196caec223786795c8730786aead6e2dbffe03403186f203075973%40group.calendar.google.com&ctz=America%2FDenver',
            'icon': 'fa-solid fa-calendar',
            'type': 'fontawesome',
        }
    ],
    'use_edit_page_button': False,
    'analytics': {'google_analytics_id': 'G-J48T2BG3J7'},
    'navbar_align': 'left',
    'navbar_start': ['navbar-logo', 'version-switcher'],
    'navbar_center': ['navbar-nav'],
    'header_links_before_dropdown': 6,
    'navbar_persistent': ['search-button'],
    'navbar_end': ['navbar-icon-links', 'theme-switcher'],
    'switcher': {
        'json_url': 'https://unidata.github.io/MetPy/pst-versions.json',
        'version_match': doc_version
    },
    'navigation_with_keys': False
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {'canonical_url': 'https://unidata.github.io/MetPy/latest/'}

# Extra variables that will be available to the templates. Used to create the
# links to the Github repository sources and issues
html_context = {
    'doc_path': 'docs',
    'galleries': sphinx_gallery_conf['gallery_dirs'],
    'gallery_dir': dict(zip(sphinx_gallery_conf['gallery_dirs'],
                            sphinx_gallery_conf['examples_dirs'])),
    'api_dir': 'api/generated',
    'github_user': 'Unidata',
    'github_repo': 'MetPy',
    'github_version': 'main',  # Make changes to the main branch
    'default_mode': 'light',
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = ' '.join((project, version))

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = str(Path('_static') / 'metpy_horizontal.png')

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = str(Path('_static') / 'metpy_32x32.ico')

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = ['theme-unidata.css']
html_js_files = ['doc_shared.js']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y at %H:%M:%S'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': ['sidebar-nav-bs']
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'MetPydoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  ('index', 'MetPy.tex', 'MetPy Documentation',
   'MetPy Developers', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'metpy', 'MetPy Documentation',
     ['MetPy Developers'], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'MetPy', 'MetPy Documentation',
   'MetPy Developers', 'MetPy', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False

# -- Options for linkcheck builder ----------------------------------------

# List of regular expressions, links matching them will be ignored.
linkcheck_ignore = [
    r'https://codecov.io/github/Unidata/MetPy',
    r'https://www\.youtube\.com/watch\?v=[\d\w\-_]+',
    r'https://youtu\.be/[\d\w\-_]+',
    # Giving 404s right now and is not going to change
    r'https://twitter\.com/MetPy',
    # AMS DOIs should be stable, but resolved link consistently 403's with linkcheck
    r'https://doi\.org/10\.1175/.*',
    # This one appears to be blocking robots
    r'https://doi\.org/10\.1088/0026-1394/45/2/004',
    # Frequently fails the linkchecker
    r'https://ams\.confex\.com/ams/[\d\w]+/meetingapp\.cgi/.*',
    # Can't seem to get around inconsistent retry errors
    r'https://doi\.org/10\.1289/ehp\.1206273',
    # Couldn't fix these 403's with user agents
    r'https://doi\.org/10\.1029/2010GL045777',
    r'https://doi\.org/10\.1098/rspa\.2004\.1430',
    # Currently giving certificate errors on GitHub
    r'https://library.wmo.int/.*',
    # For some reason GHA gets a 403 from Stack Overflow
    r'https://stackoverflow.com/questions/tagged/metpy'
    ]

# Dictionary of URL redirects allowed
linkcheck_allowed_redirects = {
    r'https://pint\.readthedocs\.io': r'https://pint\.readthedocs\.io/en/stable/',
    r'https://conda.io/docs/': r'https://conda.io/projects/conda/en/latest/',
    r'https://github.com/Unidata/MetPy/issues/new/choose': r'https://github.com/login.*choose',
    r'https://doi.org/.*': r'https://.*',
    r'https://gitter.im/Unidata/MetPy': r'https://app.gitter.im/.*MetPy.*',
    r'https://library.wmo.int/idurl/.*': r'https://library.wmo.int/.*'
}

# Domain-specific HTTP headers for requests
linkcheck_request_headers = {
    r'https://docs.github.com/': {'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux i686; '
                                                'rv:24.0) Gecko/20100101 Firefox/24.0'}
}

# Function to resolve source code links for `linkcode`
# adapted from NumPy, Pandas implementations
def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        try:  # property
            fn = inspect.getsourcefile(inspect.unwrap(obj.fget))
        except (AttributeError, TypeError):
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except TypeError:
        try:  # property
            source, lineno = inspect.getsourcelines(obj.fget)
        except (AttributeError, TypeError):
            lineno = None
    except OSError:
        lineno = None

    if lineno:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(metpy.__file__))

    if "+" in metpy.__version__:
        return f"https://github.com/Unidata/MetPy/blob/main/src/metpy/{fn}{linespec}"
    else:
        return (
            f"https://github.com/Unidata/MetPy/blob/"
            f"v{metpy.__version__}/src/metpy/{fn}{linespec}"
        )
